"""
pyart.retrieve.radar
======================

Retrievals from spectral data.

.. autosummary::
    :toctree: generated/

    compute_spectra
    compute_pol_variables_iq
    compute_reflectivity_iq
    compute_st1_iq
    compute_st2_iq
    compute_wbn_iq
    compute_differential_reflectivity_iq
    compute_mean_phase_iq
    compute_differential_phase_iq
    compute_rhohv_iq
    compute_Doppler_velocity_iq
    compute_Doppler_width_iq
    _compute_power
    _compute_autocorrelation
    _compute_lag_diff
    _compute_crosscorrelation

"""
from copy import deepcopy
from warnings import warn

import numpy as np
from scipy.signal.windows import get_window
from scipy.constants import speed_of_light

from ..config import get_metadata, get_field_name
from ..util import radar_from_spectra


def compute_spectra(radar, fields_in_list, fields_out_list, window=None):
    """
    Computes the spectra from IQ data through a Fourier transform

    Parameters
    ----------
    radar : radar object
        Object containing the IQ data
    fields_in_list : list of str
        list of input IQ data fields names
    fields_out_list : list of str
        list with the output spectra fields names obtained from the input
        fields
    window : string, tupple or None
        Parameters of the window used to obtain the spectra. The
        parameters are the ones corresponding to function
        scipy.signal.windows.get_window. If None no window will be used

    Returns
    -------
    spectra : spectra radar object
        radar object containing the spectra fields

    """
    spectra = deepcopy(radar)
    spectra.fields = {}
    for field_name_in, field_name_out in zip(fields_in_list, fields_out_list):
        if field_name_in in ('IQ_hh_ADU', 'IQ_vv_ADU'):
            spectrum = np.ma.masked_all(
                (radar.nrays, radar.ngates, radar.npulses_max),
                dtype=np.complex64)
            for ray, npuls in enumerate(radar.npulses['data']):
                wind_data = radar.fields[field_name_in]['data'][
                    ray, :, 0:npuls].filled(0.)
                if window is not None:
                    wind = get_window(window, npuls)
                    wind = wind/np.sqrt(np.sum(np.power(wind, 2.))/npuls)
                    wind = np.broadcast_to(
                        np.atleast_2d(wind), (radar.ngates, npuls))
                    wind_data *= wind

                spectrum[ray, :, 0:npuls] = np.fft.fftshift(
                    np.fft.fft(wind_data, axis=-1)/npuls, axes=-1)
        else:
            spectrum = np.ma.masked_all(
                (radar.nrays, radar.ngates, radar.npulses_max),
                dtype=np.float32)
            for ray, npuls in enumerate(radar.npulses['data']):
                spectrum[ray, :, 0:npuls] = radar.fields[field_name_in]['data'][
                    ray, :, 0:npuls]/npuls

        field_dict = get_metadata(field_name_out)
        field_dict['data'] = spectrum
        spectra.fields.update({field_name_out: field_dict})

    return spectra


def compute_pol_variables_iq(radar, fields_list, subtract_noise=False, lag=0,
                             direction='negative_away', phase_offset=0.,
                             signal_h_field=None, signal_v_field=None,
                             noise_h_field=None, noise_v_field=None):
    """
    Computes the polarimetric variables from the IQ signals in ADU

    Parameters
    ----------
    radar : IQ radar object
        Object containing the required fields
    fields_list : list of str
        list of fields to compute
    subtract_noise : Bool
        If True noise will be subtracted from the signals
    lag : int
        The time lag to use in the estimators
    direction : str
        The convention used in the Doppler mean field. Can be negative_away or
        negative_towards
    phase_offset : float. Dataset keyword
        The system differential phase offset to remove
    signal_h_field, signal_v_field, noise_h_field, noise_v_field : str
        Name of the fields in radar which contains the signal and noise.
        None will use the default field name in the Py-ART configuration file.

    Returns
    -------
    radar : radar object
        Object containing the computed fields

    """
    if signal_h_field is None:
        signal_h_field = get_field_name('IQ_hh_ADU')
    if signal_v_field is None:
        signal_v_field = get_field_name('IQ_vv_ADU')
    if noise_h_field is None:
        noise_h_field = get_field_name('IQ_noiseADU_hh')
    if noise_v_field is None:
        noise_v_field = get_field_name('IQ_noiseADU_vv')

    if ('reflectivity' in fields_list or
            'differential_reflectivity' in fields_list or
            ('cross_correlation_ratio' in fields_list and lag == 0) or
            ('spectrum_width' in fields_list and lag == 0)):
        noise_h = None
        if noise_h_field in radar.fields:
            noise_h = radar.fields[noise_h_field]['data']
        pwr_h = _compute_power(
            radar.fields[signal_h_field]['data'], noise=noise_h,
            subtract_noise=subtract_noise)

    if ('reflectivity_vv' in fields_list or
            'differential_reflectivity' in fields_list or
            ('cross_correlation_ratio' in fields_list and lag == 0) or
            ('spectrum_width_vv' in fields_list and lag == 0)):
        noise_v = None
        if noise_v_field in radar.fields:
            noise_v = radar.fields[noise_v_field]['data']
        pwr_v = _compute_power(
            radar.fields[signal_v_field]['data'], noise=noise_v,
            subtract_noise=subtract_noise)

    if ('velocity' in fields_list or 'spectrum_width' in fields_list or
            ('differential_reflectivity' in fields_list and lag == 1) or
            ('cross_correlation_ratio' in fields_list and lag == 1)):
        rlag1_h = _compute_autocorrelation(radar, signal_h_field, lag=1)
        if ('differential_reflectivity' in fields_list or
                'cross_correlation_ratio' in fields_list):
            rlagn_h_abs = np.ma.abs(rlag1_h)

    if ('velocity_vv' in fields_list or 'spectrum_width_vv' in fields_list or
            ('differential_reflectivity' in fields_list and lag == 1) or
            ('cross_correlation_ratio' in fields_list and lag == 1)):
        rlag1_v = _compute_autocorrelation(radar, signal_v_field, lag=1)
        if ('differential_reflectivity' in fields_list or
                'cross_correlation_ratio' in fields_list):
            rlagn_v_abs = np.ma.abs(rlag1_v)

    if 'spectrum_width' in fields_list and lag >= 1:
        rlag2_h = _compute_autocorrelation(radar, signal_h_field, lag=2)

    if 'spectrum_width_vv' in fields_list and lag >= 1:
        rlag2_v = _compute_autocorrelation(radar, signal_v_field, lag=2)

    if lag > 0 and ('differential_reflectivity' in fields_list or
                    'cross_correlation_ratio' in fields_list):
        rlagn_h_abs = np.ma.abs(
            _compute_autocorrelation(radar, signal_h_field, lag=1))
        rlagn_v_abs = np.ma.abs(
            _compute_autocorrelation(radar, signal_v_field, lag=1))


    if ('reflectivity' in fields_list or
            'differential_reflectivity' in fields_list):
        dBADU2dBm_h = radar.radar_calibration['dBADU_to_dBm_hh']['data'][0]
        radconst_h = (
            radar.radar_calibration['calibration_constant_hh']['data'][0])
        mfloss_h = radar.radar_calibration['matched_filter_loss_h']['data'][0]

    if ('reflectivity_vv' in fields_list or
            'differential_reflectivity' in fields_list):
        dBADU2dBm_v = radar.radar_calibration['dBADU_to_dBm_vv']['data'][0]
        radconst_v = (
            radar.radar_calibration['calibration_constant_vv']['data'][0])
        mfloss_v = radar.radar_calibration['matched_filter_loss_v']['data'][0]

    if 'reflectivity' in fields_list or 'reflectivity_vv' in fields_list:
        pathatt = radar.radar_calibration['path_attenuation']['data'][0]
        rangeKm = np.broadcast_to(
            np.atleast_2d(radar.range['data']/1000.),
            (radar.nrays, radar.ngates))

    if ('velocity' in fields_list or 'velocity_vv' in fields_list or
            'spectrum_width' in fields_list or
            'spectrum_width_vv' in fields_list):
        prt = np.broadcast_to(
            np.expand_dims(radar.instrument_parameters['prt']['data'], axis=1),
            (radar.nrays, radar.ngates))
        wavelength = (
            speed_of_light/radar.instrument_parameters['frequency']['data'][0])

    fields = {}
    if ('reflectivity' in fields_list or
            ('differential_reflectivity' in fields_list and lag == 0)):
        dBZ = (
            10.*np.ma.log10(pwr_h)+dBADU2dBm_h+radconst_h+mfloss_h +
            pathatt*rangeKm+20.*np.log10(rangeKm))

        if 'reflectivity' in fields_list:
            dBZ_field = 'reflectivity'
            dBZ_dict = get_metadata(dBZ_field)
            dBZ_dict['data'] = dBZ
            fields.update({dBZ_field: dBZ_dict})

    if ('reflectivity_vv' in fields_list or
            ('differential_reflectivity' in fields_list and lag == 0)):
        dBZv = (
            10.*np.ma.log10(pwr_v)+dBADU2dBm_v+radconst_v+mfloss_v +
            pathatt*rangeKm+20.*np.log10(rangeKm))
        if 'reflectivity_vv' in fields_list:
            dBZv_field = 'reflectivity_vv'
            dBZv_dict = get_metadata(dBZv_field)
            dBZv_dict['data'] = dBZv
            fields.update({dBZv_field: dBZv_dict})

    if 'differential_reflectivity' in fields_list:
        if lag == 0:
            zdr = dBZ-dBZv
        else:
            zdr = (
                (10.*np.ma.log10(rlagn_h_abs)+dBADU2dBm_h+radconst_h) -
                (10.*np.ma.log10(rlagn_v_abs)+dBADU2dBm_v+radconst_v))
        zdr_field = 'differential_reflectivity'
        zdr_dict = get_metadata(zdr_field)
        zdr_dict['data'] = zdr
        fields.update({zdr_field: zdr_dict})

    if 'cross_correlation_ratio' in fields_list:
        if lag == 0:
            rhohv = np.ma.abs(np.ma.mean(
                radar.fields[signal_h_field]['data'] *
                np.ma.conjugate(radar.fields[signal_v_field]['data']),
                axis=-1))
            rhohv /= np.ma.sqrt(pwr_h*pwr_v)
        else:
            rhohv = np.ma.abs(_compute_crosscorrelation(
                radar, signal_h_field, signal_v_field, lag=lag))
            rhohv /= np.ma.sqrt(rlagn_h_abs*rlagn_v_abs)
        rhohv_field = 'cross_correlation_ratio'
        rhohv_dict = get_metadata(rhohv_field)
        rhohv_dict['data'] = rhohv
        fields.update({rhohv_field: rhohv_dict})

    if 'velocity' in fields_list:
        mean_vel = np.ma.angle(rlag1_h)/(2*np.pi*prt)*(wavelength/2.)
        if direction == 'negative_away':
            mean_vel = -mean_vel
        vel_field = 'velocity'
        vel_dict = get_metadata(vel_field)
        vel_dict['data'] = mean_vel
        fields.update({vel_field: vel_dict})

    if 'velocity_vv' in fields_list:
        mean_vel = np.ma.angle(rlag1_v)/(2*np.pi*prt)*(wavelength/2.)
        if direction == 'negative_away':
            mean_vel = -mean_vel
        vel_field = 'velocity_vv'
        vel_dict = get_metadata(vel_field)
        vel_dict['data'] = mean_vel
        fields.update({vel_field: vel_dict})

    if 'spectrum_width' in fields_list:
        if lag == 0:
            width = (
                np.ma.sqrt(np.ma.log(np.ma.abs(pwr_h)/np.ma.abs(rlag1_h))) /
                (np.sqrt(2.)*np.pi*prt)*(wavelength/2.))
        else:
            width = (
                np.ma.sqrt(np.ma.log(np.ma.abs(rlag1_h)/np.ma.abs(rlag2_h))) /
                (np.sqrt(6.)*np.pi*prt)*(wavelength/2.))
        width_field = 'spectrum_width'
        width_dict = get_metadata(width_field)
        width_dict['data'] = width
        fields.update({width_field: width_dict})

    if 'spectrum_width_vv' in fields_list:
        if lag == 0:
            width = (
                np.ma.sqrt(np.ma.log(np.ma.abs(pwr_v)/np.ma.abs(rlag1_v))) /
                (np.sqrt(2.)*np.pi*prt)*(wavelength/2.))
        else:
            width = (
                np.ma.sqrt(np.ma.log(np.ma.abs(rlag1_v)/np.ma.abs(rlag2_v))) /
                (np.sqrt(6.)*np.pi*prt)*(wavelength/2.))
        width_field = 'spectrum_width_vv'
        width_dict = get_metadata(width_field)
        width_dict['data'] = width
        fields.update({width_field: width_dict})

    if 'uncorrected_differential_phase' in fields_list:
        phidp_dict = compute_differential_phase_iq(
            radar, phase_offset=phase_offset, signal_h_field=signal_h_field,
            signal_v_field=signal_v_field)
        fields.update({'uncorrected_differential_phase': phidp_dict})

    radar = radar_from_spectra(radar)

    for field_name in fields_list:
        radar.add_field(field_name, fields[field_name])

    return radar


def compute_reflectivity_iq(radar, subtract_noise=False, signal_field=None,
                            noise_field=None):
    """
    Computes the reflectivity from the IQ signal data

    Parameters
    ----------
    radar : IQ radar object
        Object containing the required fields
    subtract_noise : Bool
        If true the noise is subtracted from the power
    signal_field, noise_field : str
        Name of the signal and noise fields. None will use the default field
        name in the Py-ART configuration file.

    Returns
    -------
    dBZ_dict : field dictionary
        Field dictionary containing the reflectivity

    """
    if signal_field is None:
        signal_field = get_field_name('IQ_hh_ADU')
    if noise_field is None:
        noise_field = get_field_name('IQ_noiseADU_hh')

    if radar.radar_calibration is None:
        raise ValueError(
            'Unable to compute reflectivity. Calibration parameters unknown')

    pol = 'hh'
    if signal_field is not None and 'vv' in signal_field:
        pol = 'vv'

    if pol == 'hh':
        if ('dBADU_to_dBm_hh' not in radar.radar_calibration or
                'calibration_constant_hh' not in radar.radar_calibration):
            raise ValueError(
                'Unable to compute reflectivity. ' +
                'Calibration parameters unknown')
        dBADU2dBm = radar.radar_calibration['dBADU_to_dBm_hh']['data'][0]
        radconst = (
            radar.radar_calibration['calibration_constant_hh']['data'][0])
    else:
        if ('dBADU_to_dBm_vv' not in radar.radar_calibration or
                'calibration_constant_vv' not in radar.radar_calibration):
            raise ValueError(
                'Unable to compute reflectivity. ' +
                'Calibration parameters unknown')
        dBADU2dBm = radar.radar_calibration['dBADU_to_dBm_vv']['data'][0]
        radconst = (
            radar.radar_calibration['calibration_constant_vv']['data'][0])

    if (pol == 'hh' and 'matched_filter_loss_h' in radar.radar_calibration):
        mfloss = radar.radar_calibration['matched_filter_loss_h']['data'][0]
    elif (pol == 'vv' and
          'matched_filter_loss_v' in radar.radar_calibration):
        mfloss = radar.radar_calibration['matched_filter_loss_v']['data'][0]
    else:
        warn('Unknown matched filter losses. Assumed 0 dB')
        mfloss = 0.

    if 'path_attenuation' in radar.radar_calibration:
        pathatt = radar.radar_calibration['path_attenuation']['data'][0]
    else:
        warn('Unknown gas path attenuation. Assumed 0 dB/km')
        pathatt = 0.

    rangeKm = np.broadcast_to(
        np.atleast_2d(radar.range['data']/1000.), (radar.nrays, radar.ngates))

    noise = None
    if noise_field in radar.fields:
        noise = radar.fields[noise_field]['data']

    pwr = _compute_power(
        radar.fields[signal_field]['data'], noise=noise,
        subtract_noise=subtract_noise)

    dBZ = (
        10.*np.ma.log10(pwr)+dBADU2dBm+radconst+mfloss+pathatt*rangeKm +
        20.*np.log10(rangeKm))

    dBZ_field = 'reflectivity_'+pol

    dBZ_dict = get_metadata(dBZ_field)
    dBZ_dict['data'] = dBZ

    return dBZ_dict


def compute_st1_iq(radar, signal_field=None):
    """
    Computes the statistical test one lag fluctuation from the horizontal or
    vertical channel IQ data

    Parameters
    ----------
    radar : IQ radar object
        Object containing the required fields
    signal_field : str
        Name of the field that contain the H or V IQ data. None will use the
        default field name in the Py-ART configuration file.

    Returns
    -------
    st1_dict : field dictionary
        Field dictionary containing the st1

    """
    if signal_field is None:
        signal_field = get_field_name('IQ_hh_ADU')

    st1 = _compute_lag_diff(radar, signal_field, is_log=True, lag=1)

    st1_field = 'stat_test_lag1'
    if 'vv' in signal_field:
        st1_field += '_vv'
    st1_dict = get_metadata(st1_field)
    st1_dict['data'] = st1

    return st1_dict


def compute_st2_iq(radar, signal_field=None):
    """
    Computes the statistical test two lag fluctuation from the horizontal or
    vertical channel IQ data

    Parameters
    ----------
    radar : IQ radar object
        Object containing the required fields
    signal_field : str
        Name of the field that contain the H or V IQ data. None will use the
        default field name in the Py-ART configuration file.

    Returns
    -------
    st2_dict : field dictionary
        Field dictionary containing the st2

    """
    if signal_field is None:
        signal_field = get_field_name('IQ_hh_ADU')

    st2 = _compute_lag_diff(radar, signal_field, is_log=True, lag=2)

    st2_field = 'stat_test_lag2'
    if 'vv' in signal_field:
        st2_field += '_vv'
    st2_dict = get_metadata(st2_field)
    st2_dict['data'] = st2

    return st2_dict


def compute_wbn_iq(radar, signal_field=None):
    """
    Computes the wide band noise from the horizontal or vertical channel IQ
    data

    Parameters
    ----------
    radar : IQ radar object
        Object containing the required fields
    signal_field : str
        Name of the field that contain the H or V IQ data. None will use the
        default field name in the Py-ART configuration file.

    Returns
    -------
    wbn_dict : field dictionary
        Field dictionary containing the wide band noise

    """
    if signal_field is None:
        signal_field = get_field_name('IQ_hh_ADU')

    rlag0 = _compute_autocorrelation(radar, signal_field, lag=0)
    rlag1 = _compute_autocorrelation(radar, signal_field, lag=1)
    wbn = 20*np.ma.log10(np.ma.abs(rlag0)/np.ma.abs(rlag1))

    wbn_field = 'wide_band_noise'
    if 'vv' in signal_field:
        wbn_field += '_vv'
    wbn_dict = get_metadata(wbn_field)
    wbn_dict['data'] = wbn

    return wbn_dict


def compute_differential_reflectivity_iq(radar, subtract_noise=False, lag=0,
                                         signal_h_field=None,
                                         signal_v_field=None,
                                         noise_h_field=None,
                                         noise_v_field=None):
    """
    Computes the differential reflectivity from the horizontal and vertical
    IQ data

    Parameters
    ----------
    radar : IQ radar object
        Object containing the required fields
    subtract_noise : Bool
        If true the noise is subtracted from the power
    lag : int
        Time lag used to compute the differential reflectivity
    signal_h_field, signal_v_field, noise_h_field, noise_v_field : str
        Name of the signal and noise fields. None will use the default field
        name in the Py-ART configuration file.

    Returns
    -------
    zdr_dict : field dictionary
        Field dictionary containing the differential reflectivity

    """
    if signal_h_field is None:
        signal_h_field = get_field_name('IQ_hh_ADU')
    if signal_v_field is None:
        signal_v_field = get_field_name('IQ_vv_ADU')
    if noise_h_field is None:
        noise_h_field = get_field_name('IQ_noiseADU_hh')
    if noise_v_field is None:
        noise_v_field = get_field_name('IQ_noiseADU_vv')

    if radar.radar_calibration is None:
        raise ValueError(
            'Unable to compute spectral reflectivity. ' +
            'Calibration parameters unknown')

    if ('dBADU_to_dBm_hh' not in radar.radar_calibration or
            'dBADU_to_dBm_vv' not in radar.radar_calibration or
            'calibration_constant_hh' not in radar.radar_calibration or
            'calibration_constant_vv' not in radar.radar_calibration):
        raise ValueError(
            'Unable to compute spectral reflectivity. ' +
            'Calibration parameters unknown')

    dBADU2dBm_h = radar.radar_calibration['dBADU_to_dBm_hh']['data'][0]
    dBADU2dBm_v = radar.radar_calibration['dBADU_to_dBm_vv']['data'][0]
    radconst_h = (
        radar.radar_calibration['calibration_constant_hh']['data'][0])
    radconst_v = (
        radar.radar_calibration['calibration_constant_vv']['data'][0])

    if lag == 0:
        noise = None
        if noise_h_field in radar.fields:
            noise = radar.fields[noise_h_field]['data']

        pwr_h = _compute_power(
            radar.fields[signal_h_field]['data'], noise=noise,
            subtract_noise=subtract_noise)

        noise = None
        if noise_v_field in radar.fields:
            noise = radar.fields[noise_v_field]['data']

        pwr_v = _compute_power(
            radar.fields[signal_v_field]['data'], noise=noise,
            subtract_noise=subtract_noise)
    else:
        pwr_h = np.ma.abs(
            _compute_autocorrelation(radar, signal_h_field, lag=lag))
        pwr_v = np.ma.abs(
            _compute_autocorrelation(radar, signal_v_field, lag=lag))

    zdr = (
        (10.*np.ma.log10(pwr_h)+dBADU2dBm_h+radconst_h) -
        (10.*np.ma.log10(pwr_v)+dBADU2dBm_v+radconst_v))

    zdr_dict = get_metadata('differential_reflectivity')
    zdr_dict['data'] = zdr

    return zdr_dict


def compute_mean_phase_iq(radar, signal_field=None):
    """
    Computes the differential phase from the horizontal or vertical channel
    IQ data

    Parameters
    ----------
    radar : IQ radar object
        Object containing the required fields
    signal_field : str
        Name of the field that contain the H or V IQ data. None will use the
        default field name in the Py-ART configuration file.

    Returns
    -------
    mph_dict : field dictionary
        Field dictionary containing the mean phase

    """
    if signal_field is None:
        signal_field = get_field_name('IQ_hh_ADU')

    mph = np.ma.mean(
        np.ma.angle(radar.fields[signal_field]['data'], deg=True), axis=-1)

    mph_field = 'mean_phase'
    if 'vv' in signal_field:
        mph_field += '_vv'
    mph_dict = get_metadata(mph_field)
    mph_dict['data'] = mph

    return mph_dict


def compute_differential_phase_iq(radar, phase_offset=0., signal_h_field=None,
                                  signal_v_field=None):
    """
    Computes the differential phase from the horizontal and vertical channels
    IQ data

    Parameters
    ----------
    radar : IQ radar object
        Object containing the required fields
    phase_offset : float
        system phase offset to add
    signal_h_field, signal_v_field : str
        Name of the fields that contain the H and V IQ data. None will use the
        default field name in the Py-ART configuration file.

    Returns
    -------
    phidp_dict : field dictionary
        Field dictionary containing the differential phase

    """
    if signal_h_field is None:
        signal_h_field = get_field_name('IQ_hh_ADU')
    if signal_v_field is None:
        signal_v_field = get_field_name('IQ_vv_ADU')

    phidp = np.ma.angle(np.ma.mean(
        radar.fields[signal_h_field]['data'] *
        np.ma.conjugate(radar.fields[signal_v_field]['data']), axis=-1),
                        deg=True)-phase_offset

    if phase_offset != 0:
        phidp = (phidp+180.)%360.-180.

    phidp_field = 'uncorrected_differential_phase'
    phidp_dict = get_metadata(phidp_field)
    phidp_dict['data'] = phidp

    return phidp_dict


def compute_rhohv_iq(radar, subtract_noise=False, lag=0, signal_h_field=None,
                     signal_v_field=None, noise_h_field=None,
                     noise_v_field=None):
    """
    Computes RhoHV from the horizontal and vertical channels IQ data

    Parameters
    ----------
    radar : IQ radar object
        Object containing the required fields
    subtract_noise : Bool
        If True noise will be subtracted from the signals
    lag : int
        Time lag used in the computation
    signal_h_field, signal_v_field, noise_h_field, noise_v_field : str
        Name of the fields in radar which contains the signal and noise.
        None will use the default field name in the Py-ART configuration file.

    Returns
    -------
    rhohv_dict : field dictionary
        Field dictionary containing the RhoHV

    """
    if signal_h_field is None:
        signal_h_field = get_field_name('IQ_hh_ADU')
    if signal_v_field is None:
        signal_v_field = get_field_name('IQ_vv_ADU')
    if noise_h_field is None:
        noise_h_field = get_field_name('IQ_noiseADU_hh')
    if noise_v_field is None:
        noise_v_field = get_field_name('IQ_noiseADU_vv')

    if lag == 0:
        rhohv = np.ma.abs(np.ma.mean(
            radar.fields[signal_h_field]['data'] *
            np.ma.conjugate(radar.fields[signal_v_field]['data']), axis=-1))

        noise = None
        if noise_h_field in radar.fields:
            noise = radar.fields[noise_h_field]['data']

        pwr_h = _compute_power(
            radar.fields[signal_h_field]['data'], noise=noise,
            subtract_noise=subtract_noise)

        noise = None
        if noise_v_field in radar.fields:
            noise = radar.fields[noise_v_field]['data']

        pwr_v = _compute_power(
            radar.fields[signal_v_field]['data'], noise=noise,
            subtract_noise=subtract_noise)
    else:
        rhohv = np.ma.abs(_compute_crosscorrelation(
            radar, signal_h_field, signal_v_field, lag=lag))
        pwr_h = np.ma.abs(
            _compute_autocorrelation(radar, signal_h_field, lag=lag))
        pwr_v = np.ma.abs(
            _compute_autocorrelation(radar, signal_v_field, lag=lag))

    rhohv /= np.ma.sqrt(pwr_h*pwr_v)

    rhohv_dict = get_metadata('cross_correlation_ratio')
    rhohv_dict['data'] = rhohv

    return rhohv_dict


def compute_Doppler_velocity_iq(radar, signal_field=None,
                                direction='negative_away'):
    """
    Computes the Doppler velocity from the IQ data

    Parameters
    ----------
    radar : IQ radar object
        Object containing the required fields
    signal_field : str
        Name of the field in the radar which contains the signal.
        None will use the default field name in the Py-ART configuration file.
    direction : str
        The convention used in the Doppler mean field. Can be negative_away or
        negative_towards

    Returns
    -------
    vel_dict : field dictionary
        Field dictionary containing the Doppler velocity

    """
    if signal_field is None:
        signal_field = get_field_name('IQ_hh_ADU')

    if radar.instrument_parameters is None:
        warn('Unable to compute Doppler velocity. No instrument information')
        return None

    if ('prt' not in radar.instrument_parameters or
            'frequency' not in radar.instrument_parameters):
        warn('Unable to compute Doppler velocity. '
             'Unknown PRT or radar frequency')
        return None

    rlag_1 = _compute_autocorrelation(radar, signal_field, lag=1)
    prt = np.broadcast_to(
        np.expand_dims(radar.instrument_parameters['prt']['data'], axis=1),
        (radar.nrays, radar.ngates))
    wavelength = (
        speed_of_light/radar.instrument_parameters['frequency']['data'][0])
    mean_vel = np.ma.angle(rlag_1)/(2*np.pi*prt)*(wavelength/2.)

    if direction == 'negative_away':
        mean_vel = -mean_vel

    vel_field = 'velocity'
    if 'vv' in signal_field:
        vel_field += '_vv'
    vel_dict = get_metadata(vel_field)
    vel_dict['data'] = mean_vel

    return vel_dict


def compute_Doppler_width_iq(radar, subtract_noise=True, signal_field=None,
                             noise_field=None, lag=1):
    """
    Computes the Doppler width from the IQ data

    Parameters
    ----------
    radar : Radar radar object
        Object containing the required fields
    subtract_noise : Bool
        If True noise will be subtracted from the signals
    lag : int
        Time lag used in the denominator of the computation
    signal_field, noise_field : str
        Name of the field in the radar which contains the signal and noise.
        None will use the default field name in the Py-ART configuration file.

    Returns
    -------
    width_dict : field dictionary
        Field dictionary containing the Doppler spectrum width

    Reference
    ---------
    lag 1 algorithm based on:
        Time-Domain Computation of Mean and Variance of Doppler Spectra
        R. C. Srivastava, A. R. Jameson, and P. H. Hildebrand
        Journal of Applied Meteorology February 1979, Vol. 18, No. 2

    """
    if signal_field is None:
        signal_field = get_field_name('IQ_hh_ADU')
    if noise_field is None:
        noise_field = get_field_name('IQ_noiseADU_hh')

    prt = np.broadcast_to(
        np.expand_dims(radar.instrument_parameters['prt']['data'], axis=1),
        (radar.nrays, radar.ngates))
    wavelength = (
        speed_of_light/radar.instrument_parameters['frequency']['data'][0])

    if lag == 0:
        noise = None
        if noise_field in radar.fields:
            noise = radar.fields[noise_field]['data']

        rlag_1 = _compute_autocorrelation(radar, signal_field, lag=1)
        pwr = _compute_power(
            radar.fields[signal_field]['data'], noise=noise,
            subtract_noise=subtract_noise)

        width = (np.ma.sqrt(np.ma.log(np.ma.abs(pwr)/np.ma.abs(rlag_1))) /
                 (np.sqrt(2.)*np.pi*prt)*(wavelength/2.))
    else:
        rlag_2 = _compute_autocorrelation(radar, signal_field, lag=2)
        rlag_1 = _compute_autocorrelation(radar, signal_field, lag=1)

        width = (np.ma.sqrt(np.ma.log(np.ma.abs(rlag_1)/np.ma.abs(rlag_2))) /
                 (np.sqrt(6.)*np.pi*prt)*(wavelength/2.))

    width_field = 'spectrum_width'
    if 'vv' in signal_field:
        width_field += '_vv'
    width_dict = get_metadata(width_field)
    width_dict['data'] = width

    return width_dict


def _compute_power(signal, noise=None, subtract_noise=False):
    """
    Compute the signal power in linear units

    Parameters
    ----------
    signal : float array
        The IQ signal
    noise : float array
        The noise power per pulse
    subtract_noise : Bool
        If True and noise not None the noise power will be subtracted from the
        signal power

    Returns
    -------
    pwr : float array
        The computed signal power

    """
    pwr = np.ma.mean(np.ma.power(np.ma.abs(signal), 2.), axis=-1)

    if subtract_noise and noise is not None:
        noise_gate = np.ma.mean(noise, axis=-1)
        pwr -= noise_gate
        pwr[pwr < 0.] = np.ma.masked

    return pwr


def _compute_autocorrelation(radar, signal_field, lag=1):
    """
    Compute the signal autocorrelation in linear units

    Parameters
    ----------
    radar : IQ radar object
        The radar object containing the fields
    signal_field : str
        The IQ signal
    lag : int
        Time lag to compute

    Returns
    -------
    rlag : float array
        The computed autocorrelation lag

    """
    rlag = np.ma.masked_all((radar.nrays, radar.ngates), dtype=np.complex64)
    for ray, npulses in enumerate(radar.npulses['data']):
        if lag >= npulses:
            warn('lag larger than number of pulses in ray')
            continue
        rlag[ray, :] = np.ma.mean(
            np.ma.conjugate(
                radar.fields[signal_field]['data'][ray, :, 0:npulses-lag]) *
            radar.fields[signal_field]['data'][ray, :, lag:npulses], axis=-1)

    return rlag


def _compute_lag_diff(radar, signal_field, is_log=True, lag=1):
    """
    Compute the signal autocorrelation in linear units

    Parameters
    ----------
    radar : IQ radar object
        The radar object containing the fields
    signal_field : str
        The IQ signal
    lag : int
        Time lag to compute

    Returns
    -------
    rlag : float array
        The computed autocorrelation lag

    """
    rlag = np.ma.masked_all((radar.nrays, radar.ngates))
    pwr = np.ma.power(np.ma.abs(radar.fields[signal_field]['data']), 2.)
    if is_log:
        pwr = 10.*np.ma.log10(pwr)
    for ray, npulses in enumerate(radar.npulses['data']):
        if lag >= npulses:
            warn('lag larger than number of pulses in ray')
            continue

        rlag[ray, :] = np.ma.mean(np.ma.abs(
            pwr[ray, :, 0:npulses-lag]-pwr[ray, :, lag:npulses]), axis=-1)

    return rlag


def _compute_crosscorrelation(radar, signal_h_field, signal_v_field, lag=1):
    """
    Compute the cross-correlation between H and V in linear units

    Parameters
    ----------
    radar : IQ radar object
        The radar object containing the fields
    signal_h_field, signal_v_field : str
        The IQ H and V signal names
    lag : int
        Time lag to compute

    Returns
    -------
    rlag : float array
        The computed cross-correlation lag

    """
    rlag = np.ma.masked_all((radar.nrays, radar.ngates), dtype=np.complex64)
    for ray, npulses in enumerate(radar.npulses['data']):
        if lag >= npulses:
            warn('lag larger than number of pulses in ray')
            continue
        rlag[ray, :] = np.ma.mean(
            radar.fields[signal_h_field]['data'][ray, :, 0:npulses-lag] *
            np.ma.conjugate(
                radar.fields[signal_v_field]['data'][ray, :, lag:npulses]),
            axis=-1)

    return rlag



""" Unit Tests for Py-ART's io/mdv_grid.py module. """

import datetime
from io import BytesIO

import numpy as np
from numpy.testing import assert_raises, assert_warns, assert_almost_equal

import pyart


class Mdv_grid_Tests(object):
    """
    Class for declaring unit tests for the io/mdv_grid.py module which
    require a temporary file which will be removed at the end of the test.
    """

    def test_write_read_target(self):
        # write and read in target grid
        original_grid = pyart.testing.make_target_grid()
        tmpfile = BytesIO()
        pyart.io.write_grid_mdv(tmpfile, original_grid)
        tmpfile.seek(0)
        grid = pyart.io.read_grid_mdv(tmpfile, file_field_names=True)

        # check the contents of the grid which was read
        self.check_target_reflectivity_field(grid)
        attrs_to_check = [
            'time', 'origin_latitude', 'origin_longitude', 'origin_altitude',
            'x', 'y', 'z']
        for attr in attrs_to_check:
            self.check_attr_dics(grid, original_grid, attr)

    @staticmethod
    def check_target_reflectivity_field(grid):
        # check the reflectivity field data
        assert 'reflectivity' in grid.fields
        fdata = grid.fields['reflectivity']['data']

        assert fdata.shape == (2, 400, 320)
        assert fdata.dtype == np.float32

        assert np.abs(fdata[0, 0, 0] - 0.) <= 0.01
        assert np.abs(fdata[0, 60, 50] - 10.) <= 0.01
        assert np.abs(fdata[0, 110, 90] - 20.) <= 0.01
        assert np.abs(fdata[0, 160, 130] - 30.) <= 0.01

        assert np.abs(fdata[1, 0, 0] - 5.) <= 0.01
        assert np.abs(fdata[1, 60, 50] - 15.) <= 0.01
        assert np.abs(fdata[1, 110, 90] - 25.) <= 0.01
        assert np.abs(fdata[1, 160, 130] - 35.) <= 0.01

    @staticmethod
    def check_attr_dics(grid1, grid2, attr_name):
        print("Checking:", attr_name)
        dic1 = getattr(grid1, attr_name)
        dic2 = getattr(grid2, attr_name)
        # check that the data and units keys are similar
        assert np.allclose(dic1['data'], dic2['data'])
        assert dic1['data'].shape == dic2['data'].shape
        assert dic1['data'].dtype == dic2['data'].dtype
        if 'units' in dic1:
            assert dic1['units'] == dic2['units']

    def test_write_read_target_delay(self):
        # write and read in the target grid with delayed field loading
        original_grid = pyart.testing.make_target_grid()
        tmpfile = BytesIO()
        pyart.io.write_grid_mdv(tmpfile, original_grid)
        tmpfile.seek(0)
        grid = pyart.io.read_grid_mdv(tmpfile, file_field_names=True,
                                      delay_field_loading=True)

        # check the contents of the grid which was read
        self.check_target_reflectivity_field(grid)
        attrs_to_check = [
            'time', 'origin_latitude', 'origin_longitude', 'origin_altitude',
            'x', 'y', 'z']
        for attr in attrs_to_check:
            self.check_attr_dics(grid, original_grid, attr)

    def test_write_read_target_modified(self):
        # write and read in target grid with modifications
        original_grid = pyart.testing.make_target_grid()
        original_grid.metadata['radar_0_lon'] = -98.1
        original_grid.metadata['radar_0_lat'] = 36.74
        original_grid.metadata['radar_0_alt'] = 300.
        original_grid.metadata['instrument_name'] = 'testtesttest'
        original_grid.fields['reflectivity']['standard_name'] = 'fieldname'
        tmpfile = BytesIO()
        pyart.io.write_grid_mdv(tmpfile, original_grid)
        tmpfile.seek(0)
        grid = pyart.io.read_grid_mdv(tmpfile, file_field_names=True)

        # check the contents of the grid which was read
        self.check_target_reflectivity_field(grid)
        attrs_to_check = [
            'time', 'origin_latitude', 'origin_longitude', 'origin_altitude',
            'x', 'y', 'z']
        for attr in attrs_to_check:
            self.check_attr_dics(grid, original_grid, attr)
        assert grid.metadata['instrument_name'] == 'testtesttest'

    def test_read_write_mask(self):
        # Read and write a grid with field which is partially masked
        original_grid = pyart.testing.make_empty_grid(
            (20, 20, 20), ((0, 1), (0, 1), (0, 1)))
        fdata = np.ma.ones((20, 20, 20), dtype=np.float32)
        fdata[0, 0, 1] = np.ma.masked
        original_grid.fields['field_one'] = {'data': fdata}
        tmpfile = BytesIO()
        pyart.io.write_grid_mdv(tmpfile, original_grid)
        tmpfile.seek(0)

        # check the new grid object
        grid = pyart.io.read_grid_mdv(tmpfile, file_field_names=True)
        assert 'field_one' in grid.fields
        fdata = grid.fields['field_one']['data']
        assert fdata.dtype == np.float32
        assert fdata.shape == (20, 20, 20)
        assert np.ma.isMA(fdata)
        assert np.ma.is_masked(fdata[0, 0, 1])
        assert not np.ma.is_masked(fdata[0, 0, 0])

    def test_raise_warning_to_many_z(self):
        # Grids greater than 122 level should raise a warning and
        # write out the first 122 levels.
        grid = pyart.testing.make_empty_grid(
            (123, 2, 2), ((0, 1), (0, 1), (0, 1)))
        grid.fields['reflectivity'] = {'data': np.zeros((123, 2, 2))}
        tmpfile = BytesIO()
        assert_warns(
            UserWarning, pyart.io.write_grid_mdv, tmpfile, grid)
        tmpfile.seek(0)
        rgrid = pyart.io.read_grid_mdv(tmpfile)
        assert len(grid.z['data']) == 123
        assert len(rgrid.z['data']) == 122

    def test_write_types(self):
        # Write various data types
        original_grid = pyart.testing.make_empty_grid(
            (2, 2, 2), ((0, 1), (0, 1), (0, 1)))

        uint8_data = np.ones((2, 2, 2), dtype=np.float32) * 100. + 5
        original_grid.fields['field_one'] = {
            'data': uint8_data,
            'scale_factor': 10.,
            'add_offset': 5.,
            '_Write_as_dtype': 'uint8',
            '_FillValue': 128, }

        uint16_data = np.ones((2, 2, 2), dtype=np.float32) * 22.
        original_grid.fields['field_two'] = {
            'data': uint16_data,
            '_Write_as_dtype': 'uint16',
            'missing_value': 256, }

        float32_data = np.ones((2, 2, 2), dtype=np.float32) * 98. + 2.
        original_grid.fields['field_three'] = {
            'data': float32_data,
            'scale_factor': 10.,
            'add_offset': 2.,
            }
        tmpfile = BytesIO()
        pyart.io.write_grid_mdv(tmpfile, original_grid)
        tmpfile.seek(0)

        # verify that the data get read in correctly
        # use delay_field_loading to prevent closing of tmpfile
        grid = pyart.io.read_grid_mdv(
            tmpfile, file_field_names=True, delay_field_loading=True)
        assert np.all(grid.fields['field_one']['data'] == 105)
        assert np.all(grid.fields['field_two']['data'] == 22)
        assert np.all(grid.fields['field_three']['data'] == 100)
        tmpfile.seek(0)

        # vertify that the data is encoded to the correct type in the file
        mdvfile = pyart.io.mdv_common.MdvFile(tmpfile)
        field_name = [h['field_name'] for h in mdvfile.field_headers]
        field_one = field_name.index('field_one')
        field_two = field_name.index('field_two')
        field_three = field_name.index('field_three')
        encoding = [h['encoding_type'] for h in mdvfile.field_headers]
        assert encoding[field_one] == pyart.io.mdv_common.ENCODING_INT8
        assert encoding[field_two] == pyart.io.mdv_common.ENCODING_INT16
        assert encoding[field_three] == pyart.io.mdv_common.ENCODING_FLOAT32
        mdvfile.close()

    def test_raise_error_on_invalid_dtype(self):
        #  Unsupported dtyped field should raise a TypeError
        grid = pyart.testing.make_empty_grid(
            (2, 2, 2), ((0, 1), (0, 1), (0, 1)))
        grid.fields['reflectivity'] = {
            'data': np.zeros((2, 2, 2), dtype=np.float32),
            '_Write_as_dtype': 'float16'}
        tmpfile = BytesIO()
        assert_raises(TypeError, pyart.io.write_grid_mdv, tmpfile, grid)


def test_time_dic_to_datetime():
    dt = pyart.io.mdv_grid._time_dic_to_datetime(
        {'data': [20], 'units': 'seconds since 1970-01-01 00:00:00'})
    assert dt == datetime.datetime(1970, 1, 1, 0, 0, 20)

    dt = pyart.io.mdv_grid._time_dic_to_datetime(
        {'data': [20], 'units': 'seconds since 1970-01-01 00:00:00',
         'calendar': 'standard'})
    assert dt == datetime.datetime(1970, 1, 1, 0, 0, 20)


def test_mdv_degree_grid():
    grid = pyart.io.read_grid_mdv(
        pyart.testing.MDV_GRID_FILE, file_field_names=True)

    assert 'refl' in grid.fields.keys()
    fdata = grid.fields['refl']['data']
    assert fdata.shape == (1, 1837, 3661)
    assert np.ma.is_masked(fdata[0, 0, 0])
    assert_almost_equal(fdata[0, 130, 2536], 20.0, 1)

    assert grid.x['units'] == 'degree_E'
    assert_almost_equal(grid.x['data'][0], -129.99, 2)
    assert grid.y['units'] == 'degree_N'
    assert_almost_equal(grid.y['data'][0], 20.01, 2)



"""
pyart.aux_io.metranet_reader
============================

Routines for putting METRANET data files into radar object.
(Used by ELDES www.eldesradar.it)

.. autosummary::
    :toctree: generated/

    read_metranet
    read_metranet_c
    read_metranet_python

"""

import os
import datetime
import platform
from warnings import warn

import numpy as np
from scipy.stats import circmean

from ..config import FileMetadata, get_fillvalue
from ..io.common import make_time_unit_str, _test_arguments
from ..core.radar import Radar
from ..exceptions import MissingOptionalDependency

from .metranet_c import Selex_Angle, get_library
from .metranet_python import read_polar as read_polar_python
from .metranet_c import read_polar as read_polar_c

from .dn_to_float import nyquist_vel
# check existence of METRANET library
try:
    METRANET_LIB = get_library(momentms=False, momentpm=True)
    if platform.system() == 'Linux':
        METRANET_LIB = get_library(momentms=True)
    _METRANETLIB_AVAILABLE = True
except:
    _METRANETLIB_AVAILABLE = False

METRANET_FIELD_NAMES = {
    'WID': 'spectrum_width',
    'VEL': 'velocity',
    'ZH': 'reflectivity',
    'ZV': 'reflectivity_vv',  # non standard name
    'ZDR': 'differential_reflectivity',
    'RHO': 'uncorrected_cross_correlation_ratio',
    'PHI': 'uncorrected_differential_phase',
    'ST1': 'stat_test_lag1',  # statistical test on lag 1 (non standard name)
    'ST2': 'stat_test_lag2',  # statistical test on lag 2 (non standard name)
    'WBN': 'wide_band_noise',  # (non standard name)
    'MPH': 'mean_phase',  # (non standard name)
    'CLT': 'clutter_exit_code',  # (non standard name)
    'ZHC': 'reflectivity_hh_clut',  # cluttered horizontal reflectivity
    'ZVC': 'reflectivity_hh_clut',  # cluttered vertical reflectivity
}

PM_MOM = ["ZH", "ZV", "ZDR", "RHO", "PHI", "VEL", "WID", "ST1", "ST2", "WBN",
          "MPH"]
PH_MOM = ["ZH", "ZV", "ZDR", "RHO", "PHI", "VEL", "WID", "ST1", "ST2", "WBN",
          "MPH", "CLT"]
PL_MOM = ["ZH", "ZV", "ZDR", "RHO", "PHI", "VEL", "WID", "ZHC", "ZVC"]

NPM_MOM = 11
NPH_MOM = 12
NPL_MOM = 9


def read_metranet(filename, field_names=None, rmax=0.,
                  additional_metadata=None, file_field_names=False,
                  exclude_fields=None, reader='C', nbytes=4, **kwargs):

    """
    Read a METRANET file.

    Parameters
    ----------
    filename : str
        Name of the METRANET file to read.
    field_names : dict, optional
        Dictionary mapping METRANET field names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    rmax : float, optional
        Maximum radar range to store in the radar object [m]. If 0 all data
        will be stored
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the MDV data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters.
    reader : str
        The reader library to use. Can be either 'C' or 'python'
    nbytes : int
        The number of bytes used to store the data in numpy arrays, e.g. if
        nbytes=4 then floats are going to be stored as np.float32

    Returns
    -------
    radar : Radar
        Radar object containing data from METRANET file.

    """
    # check if it is the right file
    bfile = os.path.basename(filename)

    supported_file = (bfile.startswith('PM') or bfile.startswith('PH') or
                      bfile.startswith('PL') or bfile.startswith('MS') or
                      bfile.startswith('MH') or bfile.startswith('ML'))

    if not supported_file:
        raise ValueError(
            'Only polar data files starting by ' +
            'PM, PH, PL, MS, MH or ML are supported')

    if reader == 'C' and _METRANETLIB_AVAILABLE:
        return read_metranet_c(
            filename, field_names, rmax, additional_metadata,
            file_field_names, exclude_fields, nbytes=nbytes, **kwargs)

    if reader == 'python':
        return read_metranet_python(
            filename, field_names, rmax, additional_metadata,
            file_field_names, exclude_fields, nbytes=nbytes, **kwargs)

    warn('Invalid reader name or C library not available,' +
         ' using python (default) instead')
    return read_metranet_python(
        filename, field_names, rmax, additional_metadata,
        file_field_names, exclude_fields, nbytes=nbytes, **kwargs)


def read_metranet_c(filename, field_names=None, rmax=0.,
                    additional_metadata=None, file_field_names=False,
                    exclude_fields=None, nbytes=4, **kwargs):
    """
    Read a METRANET file.

    Parameters
    ----------
    filename : str
        Name of the METRANET file to read.
    field_names : dict, optional
        Dictionary mapping METRANET field names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    rmax : float, optional
        Maximum radar range to store in the radar object [m]. If 0 all data
        will be stored
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the MDV data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters.
    nbytes : int
        The number of bytes used to store the data in numpy arrays, e.g. if
        nbytes=4 then floats are going to be stored as np.float32

    Returns
    -------
    radar : Radar
        Radar object containing data from METRANET file.

    """
    # check that METRANET library is available
    if not _METRANETLIB_AVAILABLE:
        raise MissingOptionalDependency(
            "Metranet library is required to use read_metranet " +
            "but is not installed")

    if nbytes == 4:
        dtype = np.float32
    elif nbytes == 8:
        dtype = np.float64
    else:
        warn('Number of bytes to store the data ('+str(nbytes) +
             ') not supported. 4 bytes will be used')
        dtype = np.float32

    # test for non empty kwargs
    _test_arguments(kwargs)

    bfile = os.path.basename(filename)

    # create metadata retrieval object
    if field_names is None:
        field_names = METRANET_FIELD_NAMES
    filemetadata = FileMetadata(
        'METRANET', field_names, additional_metadata, file_field_names,
        exclude_fields)

    # get definitions from filemetadata class
    latitude = filemetadata('latitude')
    longitude = filemetadata('longitude')
    altitude = filemetadata('altitude')
    metadata = filemetadata('metadata')
    sweep_start_ray_index = filemetadata('sweep_start_ray_index')
    sweep_end_ray_index = filemetadata('sweep_end_ray_index')
    sweep_number = filemetadata('sweep_number')
    sweep_mode = filemetadata('sweep_mode')
    fixed_angle = filemetadata('fixed_angle')
    elevation = filemetadata('elevation')
    _range = filemetadata('range')
    azimuth = filemetadata('azimuth')
    _time = filemetadata('time')

    # other metadata
    frequency = filemetadata('frequency')
    beamwidth_h = filemetadata('radar_beam_width_h')
    beamwidth_v = filemetadata('radar_beam_width_v')
    pulse_width = filemetadata('pulse_width')
    rays_are_indexed = filemetadata('rays_are_indexed')
    ray_angle_res = filemetadata('ray_angle_res')
    nyquist_velocity = filemetadata('nyquist_velocity')
    npulses = filemetadata('number_of_pulses')

    ret = read_polar_c(filename, 'ZH', physic_value=True, masked_array=True)
    if ret is None:
        raise ValueError('Unable to read file '+filename)

    # total number of rays composing the sweep
    total_record = ret.header['row']
    if total_record == 0:
        raise ValueError('Number of rays in file=0.')

    # M files returning 0 pulse width. Hardcode it for the moment
    # pulse_width['data'] = np.array(
    #    [ret.header['PulseWidth']*1e-6], dtype='float64')
    pulse_width['data'] = 0.5e-6*np.ones(total_record, dtype=dtype)
    rays_are_indexed['data'] = np.array(['true'])
    ray_angle_res['data'] = np.array([1.], dtype=dtype)

    # number of gates in a ray
    num_gates = ret.header['column']

    # sweep_number (is the sweep index)
    # current sweep number (from 0 to 19)
    sweep_number['data'] = np.array([ret.header['CurrentSweep']-1])

    az_data = np.empty(total_record, dtype=dtype)
    el_data = np.empty(total_record, dtype=dtype)
    time_data = np.empty(total_record, dtype=dtype)
    ray_index_data = np.empty(total_record, dtype=dtype)

    angres = ray_angle_res['data'][0]
    valid_rays = np.ones((total_record)).astype(bool)

    ant_mode = ret.header['AntMode']  # scanning mode code
    if ant_mode == 0:
        scan_type = 'ppi'
        sweep_mode['data'] = np.array(['azimuth_surveillance'])
        # ray starting elevation angle information
        fixed_angle['data'] = np.array(
            [Selex_Angle(ret.pol_header[0].start_angle).el],
            dtype=dtype)

        # azimuth
        for i in range(total_record):
            # ray starting azimuth angle information
            if ret.pol_header[i].start_angle != 0:
                # ray starting azimuth angle information
                start_angle = Selex_Angle(ret.pol_header[i].start_angle).az
                # ray ending azimuth angle information
                end_angle = Selex_Angle(ret.pol_header[i].end_angle).az
                if end_angle > start_angle:
                    az_data[i] = start_angle + (end_angle-start_angle)/2.
                else:
                    az_data[i] = start_angle + (end_angle+360.-start_angle)/2.
                if az_data[i] > 360: # Can happen in spurious cases
                    az_data[i] -= 360
            else:
                valid_rays[i] = 0

        if not np.all(valid_rays):
            # incomplete scan

            az_full_scan = np.arange(0+angres/2, 360+angres/2, angres)
            az_closest = angres*np.floor(az_data[valid_rays]/angres)

            idx_az = np.zeros((sum(valid_rays))).astype(int)
            for i, _ in enumerate(idx_az):
                idx_az[i] = np.searchsorted(az_full_scan, az_closest[i])

            corr_az = az_full_scan
            corr_az[idx_az] = az_data[valid_rays]
            az_data = corr_az

        azimuth['data'] = az_data

        # elevation
        elevation['data'] = np.repeat(fixed_angle['data'], total_record)
    elif ant_mode == 1:
        scan_type = 'rhi'
        sweep_mode['data'] = np.array(['elevation_surveillance'])
        # ray starting azimuth angle information
        fixed_angle['data'] = np.array(
            [Selex_Angle(ret.pol_header[0].start_angle).az],
            dtype=dtype)

        # azimuth
        azimuth['data'] = np.repeat(fixed_angle['data'], total_record)

        # elevation
        for i in range(total_record):
            # ray starting elevation angle information
            start_angle = Selex_Angle(ret.pol_header[i].start_angle).el
            # ray ending elevation angle information
            end_angle = Selex_Angle(ret.pol_header[i].end_angle).el
            el_data[i] = start_angle + (end_angle-start_angle)/2.
        elevation['data'] = el_data
    elif ant_mode == 2:
        scan_type = 'sector'  # sector scan
        sweep_mode['data'] = np.array(['sector'])
        # ray starting elevation angle information
        fixed_angle['data'] = np.array(
            [Selex_Angle(ret.pol_header[0].start_angle).el],
            dtype=dtype)

        # azimuth
        for i in range(total_record):
            # ray starting azimuth angle information
            start_angle = Selex_Angle(ret.pol_header[i].start_angle).az
            # ray ending azimuth angle information
            end_angle = Selex_Angle(ret.pol_header[i].end_angle).az
            if end_angle > start_angle:
                az_data[i] = start_angle + (end_angle-start_angle)/2.
            else:
                az_data[i] = start_angle + (end_angle+360.-start_angle)/2.
        azimuth['data'] = az_data

        # elevation
        elevation['data'] = np.repeat(fixed_angle['data'], total_record)
    elif ant_mode == 3:
        scan_type = 'other'  # point of interest scan
        sweep_mode['data'] = np.array(['pointing'])
        # ray starting elevation angle information
        fixed_angle['data'] = np.array(
            [Selex_Angle(ret.pol_header[0].start_angle).el],
            dtype=dtype)

        # azimuth
        # ray starting elevation angle information
        azimuth['data'] = Selex_Angle(ret.pol_header[0].start_angle).az

        # elevation
        elevation['data'] = fixed_angle['data']
    elif ant_mode == 4:
        scan_type = 'other'  # off
        sweep_mode['data'] = np.array(['idle'])
    else:
        raise ValueError('Unknown scan type')

    if np.all(valid_rays):
        idx_az = np.arange(0, 360/angres).astype(int)

    # range (to center of beam [m])
    # distance to start of first range gate [usually 0 m]
    start_range = float(ret.header['StartRange'])
    # range resolution [m]
    gate_width = float(ret.header['GateWidth'])*1000.
    _range['data'] = np.linspace(
        start_range+gate_width/2., float(num_gates-1.) *
        gate_width+gate_width/2., num_gates, dtype=dtype)

    if rmax > 0.:
        _range['data'] = _range['data'][_range['data'] < rmax]
        nrange = len(_range['data'])

    # time (according to default_config this is the Time at the center of
    # each ray, in fractional seconds since the volume started)
    # here we find the time of end of ray since the first ray in the sweep
    for i in range(total_record):
        # time when the ray was created [s from 1.1.1970].
        # (last received pulse+processing time)
        data_time = float(ret.pol_header[i].data_time)
        # the hundreths of seconds to add to the data_time
        data_time_residue = float(ret.pol_header[i].data_time_residue)
        time_data[i] = data_time+data_time_residue/100.
        ray_index_data[i] = ret.pol_header[i].sequence

    sweep_start = min(time_data)
    start_time = datetime.datetime.utcfromtimestamp(sweep_start)
    _time['data'] = time_data-sweep_start
    _time['units'] = make_time_unit_str(start_time)

    # sweep_start_ray_index, sweep_end_ray_index
    # should be specified since start of volume but we do not have this
    # information so we specify it since start of sweep instead.
    if np.all(valid_rays):
        sweep_start_ray_index['data'] = np.array(
            [min(ray_index_data)], dtype='int32')  # ray index of first ray
        sweep_end_ray_index['data'] = np.array(
            [max(ray_index_data)], dtype='int32')   # ray index of last ray
    else:
        sweep_start_ray_index['data'] = np.array(
            [0], dtype='int32')  # ray index of first ray
        sweep_end_ray_index['data'] = np.array(
            [len(ray_index_data)], dtype='int32')   # ray index of last ray

    # ----  other information that can be obtained from metadata in file
    #       sweep information:
    #       total number of sweeps compositing the volume (i.e. 20):
    #       total_sweep=ret.pol_header[0].total_sweep
    #       total number of data bytes in the ray
    #   (num_gates*number_of_moments*(number_of_bytes in each moment)):
    #       data_bytes=ret.pol_header[0].data_bytes
    #       # time period of repetition of the volume scan:
    #       repeat_time=ret.pol_header[0].repeat_time
    #       # Nyquist velocity [m/s]:
    #       ny_quest=ret.pol_header[0].ny_quest
    #       Maximum Doppler spectrum width [m/s]:
    #       w_ny_quest=ret.pol_header[0].w_ny_quest
    #
    #       # ray specific information
    #       0 no end of sweep, 1 end of sweep, 2 end of volume scan:
    #       end_of_sweep=ret.pol_header[0].end_of_sweep
    # ------------------------------------------------------------------

    # metadata
    # get radar id
    if isinstance(ret.header["RadarName"], str):
        radar_id = ret.header["RadarName"]
    else:
        radar_id = ret.header["RadarName"].decode('utf-8')

    metadata['instrument_name'] = radar_id

    # hardcoded radar dependent metadata
    latitude['data'] = np.array([ret.header['RadarLat']], dtype=dtype)
    longitude['data'] = np.array([ret.header['RadarLon']], dtype=dtype)
    altitude['data'] = np.array([ret.header['RadarHeight']], dtype=dtype)
    frequency['data'] = np.array([ret.header['Frequency']], dtype=dtype)
    beamwidth_h['data'] = np.array([1.0], dtype=dtype)
    beamwidth_v['data'] = np.array([1.0], dtype=dtype)

    # Nyquist velocity (+-nv_value)
    nv_value = nyquist_vel(sweep_number['data'][0])
    nyquist_velocity['data'] = nv_value*np.ones(total_record, dtype=dtype)

    # number of pulses per ray
    npulses_list = []
    for i in range(total_record):
        npulses_list.append(ret.pol_header[i].pulses)
    npulses['data'] = np.array(npulses_list, dtype=np.int16)

    # fields
    fields = {}

    # ZH field
    field_name = filemetadata.get_field_name('ZH')
    if field_name is not None:
        # create field dictionary
        field_dic = filemetadata(field_name)
        # incomplete scan
        if not np.all(valid_rays):
            data = np.ma.masked_all(ret.data.shape)
            data[idx_az, :] = ret.data[idx_az, :]
        else:
            data = ret.data
        field_dic['data'] = data
        if rmax > 0:
            field_dic['data'] = field_dic['data'][:, :nrange]
        field_dic['_FillValue'] = get_fillvalue()
        fields[field_name] = field_dic

    # rest of fields
    if bfile.startswith('PM') or bfile.startswith('MS'):
        for i in range(1, NPM_MOM):
            field_name = filemetadata.get_field_name(PM_MOM[i])
            if field_name is not None:
                ret = read_polar_c(
                    filename, PM_MOM[i], physic_value=True, masked_array=True)
                if ret is None:
                    warn('Unable to read moment '+PM_MOM[i]+' in file '+filename)
                    continue
                # create field dictionary
                field_dic = filemetadata(field_name)
                if not np.all(valid_rays):
                    # incomplete scan
                    data = np.ma.masked_all(ret.data.shape)
                    data[idx_az, :] = ret.data[idx_az, :]
                else:
                    data = ret.data
                field_dic['data'] = data
                if rmax > 0:
                    field_dic['data'] = field_dic['data'][:, :nrange]
                field_dic['_FillValue'] = get_fillvalue()
                fields[field_name] = field_dic
    elif bfile.startswith('PH') or bfile.startswith('MH'):
        for i in range(1, NPH_MOM):
            field_name = filemetadata.get_field_name(PH_MOM[i])
            if field_name is not None:
                ret = read_polar_c(
                    filename, PH_MOM[i], physic_value=True, masked_array=True)
                if ret is None:
                    warn('Unable to read moment '+PH_MOM[i]+' in file '+filename)
                    continue
                # create field dictionary
                field_dic = filemetadata(field_name)
                if not np.all(valid_rays):
                    data = np.ma.masked_all(ret.data.shape)
                    data[idx_az, :] = ret.data[idx_az, :]
                else:
                    data = ret.data
                field_dic['data'] = data
                if rmax > 0:
                    field_dic['data'] = field_dic['data'][:, :nrange]
                field_dic['_FillValue'] = get_fillvalue()
                fields[field_name] = field_dic
    else:
        for i in range(1, NPL_MOM):
            field_name = filemetadata.get_field_name(PL_MOM[i])
            if field_name is not None:
                ret = read_polar_c(
                    filename, PL_MOM[i], physic_value=True, masked_array=True)
                if ret is None:
                    warn('Unable to read moment '+PL_MOM[i]+' in file '+filename)
                    continue

                # create field dictionary
                field_dic = filemetadata(field_name)
                if not np.all(valid_rays):
                    data = np.ma.masked_all(ret.data.shape)
                    data[idx_az, :] = ret.data[idx_az, :]
                else:
                    data = ret.data
                field_dic['data'] = data
                if rmax > 0:
                    field_dic['data'] = field_dic['data'][:, :nrange]
                field_dic['_FillValue'] = get_fillvalue()
                fields[field_name] = field_dic

    if not fields:
        raise ValueError('No valid moments found in '+filename)

    # instrument_parameters
    instrument_parameters = dict()
    instrument_parameters.update({'frequency': frequency})
    instrument_parameters.update({'radar_beam_width_h': beamwidth_h})
    instrument_parameters.update({'radar_beam_width_v': beamwidth_v})
    instrument_parameters.update({'pulse_width': pulse_width})
    instrument_parameters.update({'nyquist_velocity': nyquist_velocity})
    instrument_parameters.update({'number_of_pulses': npulses})

    return Radar(_time, _range, fields, metadata, scan_type, latitude,
                 longitude, altitude, sweep_number, sweep_mode, fixed_angle,
                 sweep_start_ray_index, sweep_end_ray_index, azimuth,
                 elevation, rays_are_indexed=rays_are_indexed,
                 ray_angle_res=ray_angle_res,
                 instrument_parameters=instrument_parameters)


def read_metranet_python(filename, field_names=None, rmax=0.,
                         additional_metadata=None, file_field_names=False,
                         exclude_fields=None, nbytes=4, **kwargs):
    """
    Read a METRANET file.

    Parameters
    ----------
    filename : str
        Name of the METRANET file to read.
    field_names : dict, optional
        Dictionary mapping METRANET field names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    rmax : float, optional
        Maximum radar range to store in the radar object [m]. If 0 all data
        will be stored
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the MDV data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters.
    nbytes : int
        The number of bytes used to store the data in numpy arrays, e.g. if
        nbytes=4 then floats are going to be stored as np.float32

    Returns
    -------
    radar : Radar
        Radar object containing data from METRANET file.

    """
    if nbytes == 4:
        dtype = np.float32
    elif nbytes == 8:
        dtype = np.float64
    else:
        warn('Number of bytes to store the data ('+str(nbytes) +
             ') not supported. 4 bytes will be used')
        dtype = np.float32

    # test for non empty kwargs
    _test_arguments(kwargs)

    bfile = os.path.basename(filename)

    # create metadata retrieval object
    if field_names is None:
        field_names = METRANET_FIELD_NAMES
    filemetadata = FileMetadata(
        'METRANET', field_names, additional_metadata, file_field_names,
        exclude_fields)

    # get definitions from filemetadata class
    latitude = filemetadata('latitude')
    longitude = filemetadata('longitude')
    altitude = filemetadata('altitude')
    metadata = filemetadata('metadata')
    sweep_start_ray_index = filemetadata('sweep_start_ray_index')
    sweep_end_ray_index = filemetadata('sweep_end_ray_index')

    sweep_number = filemetadata('sweep_number')
    sweep_mode = filemetadata('sweep_mode')
    fixed_angle = filemetadata('fixed_angle')
    elevation = filemetadata('elevation')
    _range = filemetadata('range')
    azimuth = filemetadata('azimuth')
    _time = filemetadata('time')

    # other metadata
    frequency = filemetadata('frequency')
    beamwidth_h = filemetadata('radar_beam_width_h')
    beamwidth_v = filemetadata('radar_beam_width_v')
    pulse_width = filemetadata('pulse_width')
    rays_are_indexed = filemetadata('rays_are_indexed')
    ray_angle_res = filemetadata('ray_angle_res')
    nyquist_velocity = filemetadata('nyquist_velocity')
    npulses = filemetadata('number_of_pulses')

    ret = read_polar_python(filename, physic_value=True, masked_array=True,
                            reorder_angles=True)
    if ret is None:
        raise ValueError('Unable to read file '+filename)

    # total number of rays composing the sweep
    total_record = ret.data['ZH'].shape[0]
    if total_record == 0:
        raise ValueError('Number of rays in file=0.')

    # M files returning 0 pulse width. Hardcode it for the moment
    # pulse_width['data'] = np.array(
    #    [ret.header['PulseWidth']*1e-6], dtype='float64')
    pulse_width['data'] = 0.5e-6*np.ones(total_record, dtype=dtype)
    rays_are_indexed['data'] = np.array(['true'])
    ray_angle_res['data'] = np.array([1.], dtype=dtype)
    angres = ray_angle_res['data'][0]

    # number of gates in a ray
    num_gates = ret.data['ZH'].shape[1]

    # sweep_number (is the sweep index)
    # current sweep number (from 0 to 19)
    sweep_number['data'] = np.array([ret.header['currentsweep']-1])

    time_data = np.empty(total_record, dtype=dtype)
    ray_index_data = np.empty(total_record, dtype=dtype)

    ant_mode = ret.header['antmode']  # scanning mode code

    no_missing_az = True # will be checked later on
    if ant_mode == 0:
        scan_type = 'ppi'
        sweep_mode['data'] = np.array(['azimuth_surveillance'])
        # ray starting elevation angle information
        fixed_angle['data'] = np.array([ret.pol_header[0]['startangle_el']],
                                       dtype=dtype)

        start_az = np.array([ray['startangle_az'] for ray in ret.pol_header])
        end_az = np.array([ray['endangle_az'] for ray in ret.pol_header])

        start_az = np.deg2rad(start_az)
        end_az = np.deg2rad(end_az)


        az_data = np.rad2deg(
            circmean(np.column_stack((start_az, end_az)), axis=1))

        if len(az_data) != 360:
            # incomplete scan

            az_full_scan = np.arange(0+angres/2, 360+angres/2, angres)
            az_closest = angres*np.floor(az_data/angres)

            idx_az = np.zeros((len(az_data))).astype(int)
            for i, _ in enumerate(idx_az):
                idx_az[i] = np.searchsorted(az_full_scan, az_closest[i])

            corr_az = az_full_scan
            corr_az[idx_az] = az_data
            az_data = corr_az
            no_missing_az = False
        azimuth['data'] = az_data

        # elevation
        elevation['data'] = np.repeat(fixed_angle['data'], total_record)
    elif ant_mode == 1:
        scan_type = 'rhi'
        sweep_mode['data'] = np.array(['elevation_surveillance'])
        # ray starting azimuth angle information
        fixed_angle['data'] = np.array([ret.pol_header[0]['startangle_az']],
                                       dtype=dtype)

        # azimuth
        azimuth['data'] = np.repeat(fixed_angle['data'], total_record)

        elevation['data'] = 0.5 * np.array([
            ray['startangle_el']+ray['endangle_el'] for ray in ret.pol_header])
    elif ant_mode == 2:
        scan_type = 'sector'  # sector scan
        sweep_mode['data'] = np.array(['sector'])
        # ray starting elevation angle information
        fixed_angle['data'] = np.array([ret.pol_header[0]['startangle_el']],
                                       dtype=dtype)

        # azimuth
        start_az = np.array([ray['startangle_az'] for ray in ret.pol_header])
        end_az = np.array([ray['endangle_az'] for ray in ret.pol_header])
        start_az = np.deg2rad(start_az)
        end_az = np.deg2rad(end_az)


        az_data = np.rad2deg(
            circmean(np.column_stack((start_az, end_az)), axis=1))

        if len(az_data) != 360:
            # incomplete scan

            az_full_scan = np.arange(0+angres/2, 360+angres/2, angres)
            az_closest = angres*np.floor(az_data/angres)

            idx_az = np.zeros((len(az_data))).astype(int)
            for i, _ in enumerate(idx_az):
                idx_az[i] = np.searchsorted(az_full_scan, az_closest[i])

            corr_az = az_full_scan
            corr_az[idx_az] = az_data
            az_data = corr_az
            no_missing_az = False

        azimuth['data'] = az_data

        # elevation
        elevation['data'] = np.repeat(fixed_angle['data'], total_record)
    elif ant_mode == 3:
        scan_type = 'other'  # point of interest scan
        sweep_mode['data'] = np.array(['pointing'])
        # ray starting elevation angle information
        fixed_angle['data'] = np.array([ret.pol_header[0]['startangle_el']],
                                       dtype=dtype)

        # azimuth
        # ray starting elevation angle information
        azimuth['data'] = np.array(
            [ret.pol_header[0]['startangle_az']], dtype=dtype)

        # elevation
        elevation['data'] = fixed_angle['data']
    elif ant_mode == 4:
        scan_type = 'other'  # off
        sweep_mode['data'] = np.array(['idle'])
    else:
        raise ValueError('Unknown scan type')

    # range (to center of beam [m])
    # distance to start of first range gate [usually 0 m]
    start_range = float(ret.header['startrange'])
    # range resolution [m]
    gate_width = float(ret.header['gatewidth'])*1000.
    _range['data'] = np.linspace(
        start_range+gate_width/2., float(num_gates-1.) *
        gate_width+gate_width/2., num_gates, dtype=dtype)

    if rmax > 0.:
        _range['data'] = _range['data'][_range['data'] < rmax]
        nrange = len(_range['data'])

    # time (according to default_config this is the Time at the center of
    # each ray, in fractional seconds since the volume started)
    # here we find the time of end of ray since the first ray in the sweep
    for i in range(total_record):
        # time when the ray was created [s from 1.1.1970].
        # (last received pulse+processing time)
        data_time = float(ret.pol_header[i]['datatime'])
        # the hundreths of seconds to add to the data_time
        data_time_residue = float(ret.pol_header[i]['datatime_residue'])
        time_data[i] = data_time+data_time_residue/100.

    ray_index_data = range(total_record)

    sweep_start = min(time_data)
    start_time = datetime.datetime.utcfromtimestamp(sweep_start)
    _time['data'] = time_data-sweep_start
    _time['units'] = make_time_unit_str(start_time)

    # sweep_start_ray_index, sweep_end_ray_index
    # should be specified since start of volume but we do not have this
    # information so we specify it since start of sweep instead.

    if no_missing_az:
        sweep_start_ray_index['data'] = np.array(
            [min(ray_index_data)], dtype='int32')  # ray index of first ray
        sweep_end_ray_index['data'] = np.array(
            [max(ray_index_data)], dtype='int32')   # ray index of last ray
    else:
        sweep_start_ray_index['data'] = np.array(
            [0], dtype='int32')  # ray index of first ray
        sweep_end_ray_index['data'] = np.array(
            [360], dtype='int32')   # ray index of last ray

    # ----  other information that can be obtained from metadata in file
    #       sweep information:
    #       total number of sweeps compositing the volume (i.e. 20):
    #       total_sweep=ret.pol_header[0].total_sweep
    #       total number of data bytes in the ray
    #   (num_gates*number_of_moments*(number_of_bytes in each moment)):
    #       data_bytes=ret.pol_header[0].data_bytes
    #       # time period of repetition of the volume scan:
    #       repeat_time=ret.pol_header[0].repeat_time
    #       # Nyquist velocity [m/s]:
    #       ny_quest=ret.pol_header[0].ny_quest
    #       Maximum Doppler spectrum width [m/s]:
    #       w_ny_quest=ret.pol_header[0].w_ny_quest
    #
    #       # ray specific information
    #       0 no end of sweep, 1 end of sweep, 2 end of volume scan:
    #       end_of_sweep=ret.pol_header[0].end_of_sweep
    #       number of pulses used in data integration:
    #       pulses=ret.pol_header[0].pulses
    # ------------------------------------------------------------------

    # metadata
    # get radar id
    if isinstance(ret.header["radarname"], str):
        radar_id = ret.header["radarname"]
    else:
        radar_id = ret.header["radarname"].decode('utf-8')

    metadata['instrument_name'] = radar_id

    # hardcoded radar dependent metadata
    latitude['data'] = np.array([ret.header['radarlat']], dtype=dtype)
    longitude['data'] = np.array([ret.header['radarlon']], dtype=dtype)
    altitude['data'] = np.array([ret.header['radarheight']], dtype=dtype)
    frequency['data'] = np.array([ret.header['frequency']], dtype=dtype)
    beamwidth_h['data'] = np.array([1.0], dtype=dtype)
    beamwidth_v['data'] = np.array([1.0], dtype=dtype)

    # Nyquist velocity (+-nv_value)
    nv_value = nyquist_vel(sweep_number['data'][0])
    nyquist_velocity['data'] = nv_value*np.ones(total_record, dtype=dtype)

    # number of pulses per ray
    if 'pulses' in ret.pol_header[0]:
        pulse_name = 'pulses'
    else:
        pulse_name = 'numpulses'

    npulses_list = []
    for i in range(total_record):
        npulses_list.append(ret.pol_header[i][pulse_name])
    npulses['data'] = np.array(npulses_list, dtype=np.int16)

    # fields
    fields = {}

    # rest of fields
    if bfile.startswith('MS'):
        nmoments = NPM_MOM
        momnames = PM_MOM
    elif bfile.startswith('MH'):
        nmoments = NPH_MOM
        momnames = PH_MOM
    else:
        nmoments = NPL_MOM
        momnames = PL_MOM

    for i in range(0, nmoments):
        field_name = filemetadata.get_field_name(momnames[i])
        if field_name is not None:

            # create field dictionary
            field_dic = filemetadata(field_name)
            if momnames[i] not in ret.data:
                warn('Moment '+momnames[i]+' not in file')
                continue
            data = ret.data[momnames[i]]

            # Check if scan is complete
            if len(data) != len(azimuth['data']):
                tmp = np.ma.masked_all((len(azimuth['data']), data.shape[1]))
                tmp[idx_az, :] = data
                data = tmp

            field_dic['data'] = data

            if rmax > 0:
                field_dic['data'] = field_dic['data'][:, :nrange]

            field_dic['_FillValue'] = get_fillvalue()
            fields[field_name] = field_dic

    if not fields:
        raise ValueError('No valid moments found in '+filename)

    # instrument_parameters
    instrument_parameters = dict()
    instrument_parameters.update({'frequency': frequency})
    instrument_parameters.update({'radar_beam_width_h': beamwidth_h})
    instrument_parameters.update({'radar_beam_width_v': beamwidth_v})
    instrument_parameters.update({'pulse_width': pulse_width})
    instrument_parameters.update({'nyquist_velocity': nyquist_velocity})
    instrument_parameters.update({'number_of_pulses': npulses})

    return Radar(_time, _range, fields, metadata, scan_type, latitude,
                 longitude, altitude, sweep_number, sweep_mode, fixed_angle,
                 sweep_start_ray_index, sweep_end_ray_index, azimuth,
                 elevation, rays_are_indexed=rays_are_indexed,
                 ray_angle_res=ray_angle_res,
                 instrument_parameters=instrument_parameters)



"""
pyart.aux_io.mf_grib_reader
==========================

Routines for putting MeteoFrance operational radar data contained in grib
files into grid object.

.. autosummary::
    :toctree: generated/

    read_png
    _get_physical_data

"""

from copy import deepcopy

import numpy as np

# check existence of pygrib library
try:
    import pygrib
    _PYGRIB_AVAILABLE = True
except ImportError:
    _PYGRIB_AVAILABLE = False

from ..config import FileMetadata
from ..io.common import _test_arguments
from ..core.grid import Grid
from ..core.transforms import geographic_to_cartesian
from ..exceptions import MissingOptionalDependency
from ..util import ma_broadcast_to

GRIB_FIELD_NAMES = {
    'hail': 'precipitation_type',
    'snow': 'precipitation_type',
    'rain': 'precipitation_type',
    'ice': 'precipitation_type'
}


def read_grib(filename, additional_metadata=None,
              field_name='precipitation_type', **kwargs):

    """
    Read a MeteoFrance data png file.

    Parameters
    ----------
    filename : str
        Name of the file to read.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
   field_name : str
        Py-ART name of the field contained in the file

    Returns
    -------
    grid : Grid
        Grid object containing the data.

    """
    # check that wradlib is available
    if not _PYGRIB_AVAILABLE:
        raise MissingOptionalDependency(
            "imageio is required to use read_grib but is not installed")

    # test for non empty kwargs
    _test_arguments(kwargs)

    grbs = pygrib.open(filename)
    for grb in grbs:
        data = grb.values
        lats = grb.distinctLatitudes
        lons = grb.distinctLongitudes
        dt_file = grb.validDate
    grbs.close()
    ny, nx = data.shape
    nz = 1

    # reserved_variables = [
    #     'time', 'x', 'y', 'z',
    #     'origin_latitude', 'origin_longitude', 'origin_altitude',
    #     'point_x', 'point_y', 'point_z', 'projection',
    #     'point_latitude', 'point_longitude', 'point_altitude',
    #     'radar_latitude', 'radar_longitude', 'radar_altitude',
    #     'radar_name', 'radar_time', 'base_time', 'time_offset',
    #     'ProjectionCoordinateSystem']

    # metadata
    filemetadata = FileMetadata('GRIB', GRIB_FIELD_NAMES, additional_metadata)

    # required reserved variables
    time = filemetadata('grid_time')
    origin_latitude = filemetadata('origin_latitude')
    origin_longitude = filemetadata('origin_longitude')
    origin_altitude = filemetadata('origin_altitude')
    x = filemetadata('x')
    y = filemetadata('y')
    z = filemetadata('z')

    # origin of webmercator
    origin_latitude['data'] = np.array([0.])
    origin_longitude['data'] = np.array([0.])
    origin_altitude['data'] = np.array([0.])

    # projection (web mercator)
    projection = {
        'proj': 'eqc',
        'R': '6371229'
    }

    x0, y0 = geographic_to_cartesian(lons.min(), lats.min(), projection)
    x1, y1 = geographic_to_cartesian(lons.max(), lats.max(), projection)
    xres = (x1-x0)/nx
    yres = (y1-y0)/ny
    x['data'] = np.arange(nx)*xres+x0
    y['data'] = np.arange(ny)*yres+y0
    z['data'] = np.array([0.])

    # Time
    time['units'] = 'seconds since '+dt_file.strftime('%Y-%m-%d 00:00:00')
    time['data'] = np.array([0])

    # read in the fields
    fields = {}
    field_dict = filemetadata(field_name)
    field_dict['data'] = ma_broadcast_to(
        _get_physical_data(data)[::-1, :], (nz, ny, nx))
    fields[field_name] = field_dict

    # radar variables
    radar_latitude = None
    radar_longitude = None
    radar_altitude = None
    radar_name = None
    radar_time = None

    return Grid(
        time, fields, dict(), origin_latitude, origin_longitude,
        origin_altitude, x, y, z, projection=projection,
        radar_latitude=radar_latitude, radar_longitude=radar_longitude,
        radar_altitude=radar_altitude, radar_name=radar_name,
        radar_time=radar_time)


def _get_physical_data(data):
    """
    gets data in physical units

    Parameters
    ----------
    data : int ndarray
        the data as int

    Returns
    -------
    data_ph : float ndarray
        the data in physical units

    """
    data_ph = deepcopy(data)

    #   0 Précipitations indétectables/Absence de neige au sol
    #   1 Précipitations indétectables/Présence de neige au sol
    #   2 Précipitations indétectables/Sol invisible (nuit ou nuages)
    #   3 Bruine
    #   4 Bruine sur sol gelé
    #   5 Bruine verglaçante
    #   6 Pluie
    #   7 Pluie sur sol gelé
    #   8 Pluie verglaçante
    #   9 Pluie et neige mêlées
    #  10 Neige mouillée
    #  11 Neige mouillée tenant au sol
    #  12 Neige humide
    #  13 Neige humide tenant au sol
    #  14 Neige sèche
    #  15 Neige sèche tenant au sol
    #  16 Granules de glace
    #  17 Grésil/Petite grêle
    #  18 Moyenne grêle
    #  19 Grosse grêle
    # 255 Indéterminé

    # No data
    data_ph[data == 255] = 20

    return data_ph



"""
pyart.io.cf1
============

Utilities for reading CF1 files.

.. autosummary::
    :toctree: generated/
    :template: dev_template.rst

    _NetCDFVariableDataExtractor

.. autosummary::
    :toctree: generated/

    read_cf1

"""

import datetime

import numpy as np
import netCDF4

from ..config import FileMetadata
from ..io.common import _test_arguments, make_time_unit_str
from ..core.radar import Radar
from ..io.cfradial import _find_all_meta_group_vars, _ncvar_to_dict
from ..io.cfradial import _unpack_variable_gate_field_dic


# Variables and dimensions in the instrument_parameter convention and
# radar_parameters sub-convention that will be read from and written to
# CfRadial files using Py-ART.
# The meta_group attribute cannot be used to identify these parameters as
# it is often set incorrectly.
_INSTRUMENT_PARAMS_DIMS = {
    # instrument_parameters sub-convention
    'frequency': ('frequency'),
    'follow_mode': ('sweep', 'string_length'),
    'pulse_width': ('time', ),
    'prt_mode': ('sweep', 'string_length'),
    'prt': ('time', ),
    'prt_ratio': ('time', ),
    'polarization_mode': ('sweep', 'string_length'),
    'nyquist_velocity': ('time', ),
    'unambiguous_range': ('time', ),
    'n_samples': ('time', ),
    'sampling_ratio': ('time', ),
    # radar_parameters sub-convention
    'radar_antenna_gain_h': (),
    'radar_antenna_gain_v': (),
    'radar_beam_width_h': (),
    'radar_beam_width_v': (),
    'radar_receiver_bandwidth': (),
    'radar_measured_transmit_power_h': ('time', ),
    'radar_measured_transmit_power_v': ('time', ),
    'radar_rx_bandwidth': (),           # non-standard
    'measured_transmit_power_v': ('time', ),    # non-standard
    'measured_transmit_power_h': ('time', ),    # non-standard
}


def read_cf1(filename, field_names=None, additional_metadata=None,
             file_field_names=False, exclude_fields=None,
             include_fields=None, delay_field_loading=False, **kwargs):
    """
    Read a CF-1 netCDF file.

    Parameters
    ----------
    filename : str
        Name of CF/Radial netCDF file to read data from.
    field_names : dict, optional
        Dictionary mapping field names in the file names to radar field names.
        Unlike other read functions, fields not in this dictionary or having a
        value of None are still included in the radar.fields dictionary, to
        exclude them use the `exclude_fields` parameter. Fields which are
        mapped by this dictionary will be renamed from key to value.
    additional_metadata : dict of dicts, optional
        This parameter is not used, it is included for uniformity.
    file_field_names : bool, optional
        True to force the use of the field names from the file in which
        case the `field_names` parameter is ignored. False will use to
        `field_names` parameter to rename fields.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.
    delay_field_loading : bool
        True to delay loading of field data from the file until the 'data'
        key in a particular field dictionary is accessed.  In this case
        the field attribute of the returned Radar object will contain
        LazyLoadDict objects not dict objects.  Delayed field loading will not
        provide any speedup in file where the number of gates vary between
        rays (ngates_vary=True) and is not recommended.

    Returns
    -------
    radar : Radar
        Radar object.

    Notes
    -----
    This function has not been tested on "stream" Cfradial files.

    """
    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrieval object
    filemetadata = FileMetadata('cfradial', field_names, additional_metadata,
                                file_field_names, exclude_fields)
    latitude = filemetadata('latitude')
    longitude = filemetadata('longitude')
    altitude = filemetadata('altitude')
    altitude_agl = filemetadata('altitude_agl')
    azimuth = filemetadata('azimuth')
    elevation = filemetadata('elevation')
    sweep_start_ray_index = filemetadata('sweep_start_ray_index')
    sweep_end_ray_index = filemetadata('sweep_end_ray_index')
    sweep_number = filemetadata('sweep_number')
    sweep_mode = filemetadata('sweep_mode')
    fixed_angle = filemetadata('fixed_angle')
    _range = filemetadata('range')
    _time = filemetadata('time')

    # read the data
    ncobj = netCDF4.Dataset(filename)
    ncvars = ncobj.variables

    # 4.1 Global attribute -> move to metadata dictionary
    metadata = dict([(k, getattr(ncobj, k)) for k in ncobj.ncattrs()])
    if 'n_gates_vary' in metadata:
        metadata['n_gates_vary'] = 'false'  # corrected below

    # 4.2 Dimensions (do nothing)

    # 4.3 Global variable -> move to metadata dictionary
    if 'volume_number' in ncvars:
        metadata['volume_number'] = int(ncvars['volume_number'][:])
    else:
        metadata['volume_number'] = 0

    global_vars = {'platform_type': 'fixed', 'instrument_type': 'radar',
                   'primary_axis': 'axis_z'}
    # ignore time_* global variables, these are calculated from the time
    # variable when the file is written.
    for var, default_value in global_vars.items():
        if var in ncvars:
            metadata[var] = str(netCDF4.chartostring(ncvars[var][:]))
        else:
            metadata[var] = default_value

    # 4.4 coordinate variables -> create attribute dictionaries
    time = _ncvar_to_dict(ncvars['time'])
    microsec = _ncvar_to_dict(ncvars['microsec'])
    _time['data'] = time['data']+1e-6*microsec['data']
    _time['data'] -= time['data'][0]
    sweep_start = (
        datetime.datetime(1970, 1, 1) +
        datetime.timedelta(seconds=int(time['data'][0])))
    _time['units'] = make_time_unit_str(sweep_start)

    _range['data'] = _ncvar_to_dict(ncvars['range'])['data']

    # 4.5 Ray dimension variables

    # 4.6 Location variables -> create attribute dictionaries
    latitude['data'] = np.array([metadata['Latitude']])
    longitude['data'] = np.array([metadata['Longitude']])
    altitude['data'] = np.array([metadata['Altitude']])
    altitude_agl['data'] = np.array([metadata.get('altitude_agl', None)])

    # 4.8 Sensor pointing variables -> create attribute dictionaries
    azimuth['data'] = _ncvar_to_dict(ncvars['azi'])['data']
    elevation['data'] = _ncvar_to_dict(ncvars['elv'])['data']

    if 'scan_rate' in ncvars:
        scan_rate = _ncvar_to_dict(ncvars['scan_rate'])
    else:
        scan_rate = None

    if 'antenna_transition' in ncvars:
        antenna_transition = _ncvar_to_dict(ncvars['antenna_transition'])
    else:
        antenna_transition = None

    # 4.7 Sweep variables -> create attribute dictionaries
    sweep_start_ray_index['data'] = np.array([0])
    sweep_end_ray_index['data'] = np.array(
        [azimuth['data'].size-1], dtype=np.int32)
    sweep_number['data'] = np.array([0])

    nazi_unique = np.unique(azimuth['data'])
    nele_unique = np.unique(elevation['data'])
    if nele_unique.size == 1 and nazi_unique.size == 1:
        scan_type = 'other'
        sweep_mode['data'] = np.array(['pointing'])
        fixed_angle['data'] = nele_unique
    elif nele_unique.size == 1 and nazi_unique.size > 1:
        scan_type = 'ppi'
        sweep_mode['data'] = np.array(['azimuth_surveillance'])
        fixed_angle['data'] = nele_unique
    elif nele_unique.size > 1 and nazi_unique.size == 1:
        scan_type = 'rhi'
        sweep_mode['data'] = np.array(['elevation_surveillance'])
        fixed_angle['data'] = nazi_unique
    else:
        raise ValueError(
            'Only single sweeps PPI, RHI or pointing are supported')

    if 'target_scan_rate' in ncvars:
        target_scan_rate = _ncvar_to_dict(ncvars['target_scan_rate'])
    else:
        target_scan_rate = None
    if 'rays_are_indexed' in ncvars:
        rays_are_indexed = _ncvar_to_dict(ncvars['rays_are_indexed'])
    else:
        rays_are_indexed = None
    if 'ray_angle_res' in ncvars:
        ray_angle_res = _ncvar_to_dict(ncvars['ray_angle_res'])
    else:
        ray_angle_res = None

    # 4.9 Moving platform geo-reference variables
    # Aircraft specific varaibles
    if 'rotation' in ncvars:
        rotation = _ncvar_to_dict(ncvars['rotation'])
    else:
        rotation = None

    if 'tilt' in ncvars:
        tilt = _ncvar_to_dict(ncvars['tilt'])
    else:
        tilt = None

    if 'roll' in ncvars:
        roll = _ncvar_to_dict(ncvars['roll'])
    else:
        roll = None

    if 'drift' in ncvars:
        drift = _ncvar_to_dict(ncvars['drift'])
    else:
        drift = None

    if 'heading' in ncvars:
        heading = _ncvar_to_dict(ncvars['heading'])
    else:
        heading = None

    if 'pitch' in ncvars:
        pitch = _ncvar_to_dict(ncvars['pitch'])
    else:
        pitch = None

    if 'georefs_applied' in ncvars:
        georefs_applied = _ncvar_to_dict(ncvars['georefs_applied'])
    else:
        georefs_applied = None

    # 4.10 Moments field data variables -> field attribute dictionary
    if 'ray_n_gates' in ncvars:
        # all variables with dimensions of n_points are fields.
        keys = [k for k, v in ncvars.items()
                if v.dimensions == ('n_points', )]
    else:
        # all variables with dimensions of 'time', 'range' are fields
        keys = [k for k, v in ncvars.items()
                if v.dimensions == ('time', 'range')]

    fields = {}
    for key in keys:
        field_name = filemetadata.get_field_name(key)
        if field_name is None:
            if exclude_fields is not None and key in exclude_fields:
                if key not in include_fields:
                    continue
            if include_fields is None or key in include_fields:
                field_name = key
            else:
                continue
        fields[field_name] = _ncvar_to_dict(ncvars[key], delay_field_loading)
        if not delay_field_loading:
            db = fields[field_name].get('db', 0)
            if db:
                # warn(field_name+' data will be converted to log')
                fields[field_name]['data'] = 10.*np.ma.log10(
                    fields[field_name]['data'])

    if 'ray_n_gates' in ncvars:
        shape = (len(ncvars['time']), len(ncvars['range']))
        ray_n_gates = ncvars['ray_n_gates'][:]
        ray_start_index = ncvars['ray_start_index'][:]
        for dic in fields.values():
            _unpack_variable_gate_field_dic(
                dic, shape, ray_n_gates, ray_start_index)

    # 4.5 instrument_parameters sub-convention -> instrument_parameters dict
    # 4.6 radar_parameters sub-convention -> instrument_parameters dict
    keys = [k for k in _INSTRUMENT_PARAMS_DIMS.keys() if k in ncvars]
    instrument_parameters = dict((k, _ncvar_to_dict(ncvars[k])) for k in keys)
    if instrument_parameters == {}:  # if no parameters set to None
        instrument_parameters = None

    # 4.7 lidar_parameters sub-convention -> skip

    # 4.8 radar_calibration sub-convention -> radar_calibration
    keys = _find_all_meta_group_vars(ncvars, 'radar_calibration')
    radar_calibration = dict((k, _ncvar_to_dict(ncvars[k])) for k in keys)
    if radar_calibration == {}:
        radar_calibration = None

    # do not close file is field loading is delayed
    if not delay_field_loading:
        ncobj.close()
    return Radar(
        _time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=instrument_parameters,
        radar_calibration=radar_calibration,
        altitude_agl=altitude_agl,
        scan_rate=scan_rate,
        antenna_transition=antenna_transition,
        target_scan_rate=target_scan_rate,
        rays_are_indexed=rays_are_indexed, ray_angle_res=ray_angle_res,
        rotation=rotation, tilt=tilt, roll=roll, drift=drift, heading=heading,
        pitch=pitch, georefs_applied=georefs_applied)



"""
pyart.retrieve.gecsx
==================

Functions for visibility and ground echoes estimation from a DEM.

"""
from copy import deepcopy
import logging
import numpy as np
from scipy.ndimage import sobel
from warnings import warn

try:
    import pyproj
    _PYPROJ_AVAILABLE = True
except ImportError:
    _PYPROJ_AVAILABLE = False
    
from ..config import get_field_name, get_metadata, get_fillvalue
from ..core import antenna_vectors_to_cartesian, antenna_to_cartesian
from . import _gecsx_functions as gf

MANDATORY_RADAR_SPECS = ['tau', 'loss', 'power', 'frequency', 'beamwidth',
                         'gain']

def gecsx(radar, radar_specs, dem_grid,
                         fill_value = None,
                         terrain_altitude_field = None,
                         bent_terrain_altitude_field = None,
                         terrain_slope_field = None,
                         terrain_aspect_field = None,
                         elevation_angle_field = None,
                         visibility_field = None,
                         min_vis_elevation_field = None,
                         min_vis_altitude_field = None,
                         incident_angle_field = None,
                         effective_area_field = None,
                         sigma_0_field = None,
                         rcs_clutter_field = None,
                         dBm_clutter_field = None,
                         dBZ_clutter_field = None,
                         visibility_polar_field = None,
                         az_conv = 0,
                         dr = 100,
                         daz = 0.012,
                         ke = 4/3.,
                         atm_att = 0.2,
                         mosotti_kw = 0.9644,
                         raster_oversampling = 1,
                         sigma0_method = 'Gabella',
                         clip = True,
                         return_pyart_objects = True,
                         verbose = True):
    """
    Estimate the radar visibility and ground clutter echoes from a digital
    elevation model

    Parameters
    ----------
    radar : radar
        Radar object which contains the coordinates on which the visibility
        will be computed
    radar_specs : dict
        Dictionary containing the radar specifications. Must have the following
        keys:
            "frequency" : radar frequenty in GHz
            "tau"       : radar pulsewidth in m
            "beamwidth" : antenna 3dB beamwidth in deg
            "loss"      : total loss (antenna + radar system) in dB
            "gain"      : total gain (antenna + radar system) in dB
            "power"     : radar total power in W
    terrain_altitude_field : str, optional
        Field name which represents the terrain altitude class field.
        A value of None will use the default field name as defined in the
        Py-ART configuration file.
    bent_terrain_altitude_field : str, optional
        Field name which represents the terrain altitude after radar
        refractivity height correction class field.
        A value of None will use the default field name as defined in the
        Py-ART configuration file.
    terrain_slope_field : str, optional
        Field name which represents the terrain slope class field.
        A value of None will use the default field name as defined in the
        Py-ART configuration file.
    terrain_aspect_field : str, optional
        Field name which represents the terrain aspect class field.
        A value of None will use the default field name as defined in the
        Py-ART configuration file.
    elevation_angle_field : str, optional
        Field name which represents the elevation angle class field.
        A value of None will use the default field name as defined in the
        Py-ART configuration file.
    visibility_field : str, optional
        Field name which represents the Cartesian visibility field.
        A value of None will use the default field name as defined in the
        Py-ART configuration file.
    min_vis_elevation_field : str, optional
        Field name which represents the minimum visible elevation angle field.
        A value of None will use the default field name as defined in the
        Py-ART configuration file.
    min_vis_altitude_field : str, optional
        Field name which represents the minimum visible altitude field.
        A value of None will use the default field name as defined in the
        Py-ART configuration file.
    sigma_0_field : str, optional
        Field name which represents the ratio between rcs and backscattering
        area field. A value of None will use the default field name as defined
        in the Py-ART configuration file.
    incident_angle_field : str, optional
        Field name which represents the incidence angle at the topography
        field. A value of None will use the default field name as defined
        in the Py-ART configuration file.
    effective_area_field : str, optional
        Field name which represents the effective backscattering area. A
        value of None will use the default field name as defined
        in the Py-ART configuration file.
    rcs_clutter_field : str, optional
        Field name which represents the backscattering cross section.
        A value of None will use the default field name as defined
        in the Py-ART configuration file.
    dBm_clutter_field : str, optional
        Field name which represents the ground clutter power signal
        field. A value of None will use the default field name as defined
        in the Py-ART configuration file.
    dBZ_clutter_field : str, optional
        Field name which represents the ground clutter reflectivity field.
        A value of None will use the default field name as defined
        in the Py-ART configuration file.
    visibility_polar_field : str, optional
        Field name which represents the polar visibility field.
        A value of None will use the default field name as defined
        in the Py-ART configuration file.
    az_conv : float, optional
        If larger than zero assumes that the antenna moves in azimuth direction
        (PPI) and averages the rx signal over the angle given by this keyword
    dr : float, optional
        Range discretization used when computing the Cartesian visibility field
        the larger the better but the slower the processing will be
    daz : float, optional
        Azimuth discretization used when computing the Cartesian visibility
        field, the larger the better but the slower the processing will be
    ke : float(optional)
        Equivalent earth-radius factor used in the computation of the radar
        beam refraction
    atm_att : float, optional
        One-way atmospheric refraction in db / km
    mosotti_kw : float, optional
        Clausius-Mosotti factor K, depends on material (water) and wavelength
        for water = sqrt(0.93)
    raster_oversampling : int, optional
        The raster resolution of the DEM should be smaller than
        the range resolution of the radar (defined by the pulse length).
        If this is not the case, this keyword can be set to increase the
        raster resolution. The values for the elevation, sigma naught,
        visibility are repeated. The other values are recalculated.
        Values for raster_oversampling:
        0 or undefined: No oversampling is done
        1: Oversampling is done. The factor N is automatically calculated
        such that 2*dx/N < pulse length
        2 or larger: Oversampling is done with this value as N
    sigma0_method : string, optional
        Which estimation method to use, either 'Gabella' or 'Delrieu'
    clip : bool, optional
        If set to true, the provided DEM will be clipped to the extent
        of the polar radar domain. Increases computation speed a lot but
        Cartesian output fields will be available only over radar domain
    return_pyart_objects : bool, optional
        If set to true, the generated products will be directly added to the
        input DEM Grid object (for Cartesian products) and the input Radar
        object (for polar products). Otherwise all products will be output
        separately
    verbose : bool, optional
        If set to true, the program will display info about the current
        progress

    Returns
    -------
    bent_terrain_altitude_dic : dict
        Terrain altitude corrected for radar beam height data and metadata
    terrain_slope_dic : dict
        Terrain slope data and metadata
    terrain_aspect_dic,: dict
        Terrain aspect data and metadata
    elevation_dic,: dict
        Elevation angle at topography data and metadata
    min_vis_elevation_dic,: dict
        Minimum visible elevation data and metadata
    min_vis_altitude_dic,: dict
        Minimum visible altitude data and metadata
    visibility_dic,: dict
        Visibility over a Cartesian domain data and metadata
    incident_angle_dic,: dict
        Incidence angle at topography data and metadata
    effective_area_dic,: dict
        Effective backscattering area data and metadata
    sigma_0_dic,: dict
        Ratio between RCS and backscattering area data and metadata
    rcs_clutter_dic,: dict
        Radar cross-section data and metadata
    dBm_clutter_dic,: dict
        Ground clutter power data and metadata
    dBZ_clutter_dic,: dict
        Ground clutter reflectivity data and metadata
    visibility_polar_dic,: dict
        Visibility over the polar radar domain data and metadata

    References
    ----------
    Gabella, M., & Perona, G. (1998). Simulation of the Orographic Influence
    on Weather Radar Using a Geometric–Optics Approach, Journal of Atmospheric
    and Oceanic Technology, 15(6), 1485-1494.
    """

    if not _PYPROJ_AVAILABLE:
        warn("gdal is required to use gecsx but is not installed")
        return None

    if verbose:
        logging.basicConfig(level = logging.INFO)
    else:
        logging.basicConfig(level = logging.WARNING)

    if radar.scan_type != 'ppi':
        raise NotImplementedError('Currently only ppi scans are supported!')

    for spec in MANDATORY_RADAR_SPECS:
        if spec not in radar_specs.keys():
            raise ValueError('Key {:s} is missing from '.format(spec) +
                             'radar_specs argument!')

    # parse fill value
    if fill_value is None:
        fill_value = get_fillvalue()

    # parse field names
    if terrain_altitude_field is None:
        terrain_altitude_field = get_field_name('terrain_altitude')
    if bent_terrain_altitude_field is None:
        bent_terrain_altitude_field = get_field_name('bent_terrain_altitude')
    if terrain_slope_field is None:
        terrain_slope_field = get_field_name('terrain_slope')
    if terrain_aspect_field is None:
        terrain_aspect_field = get_field_name('terrain_aspect')
    if elevation_angle_field is None:
        elevation_angle_field = get_field_name('elevation_angle')
    if visibility_field is None:
        visibility_field = get_field_name('visibility')
    if min_vis_elevation_field is None:
        min_vis_elevation_field = get_field_name('min_vis_elevation')
    if min_vis_altitude_field is None:
        min_vis_altitude_field = get_field_name('min_vis_altitude')
    if incident_angle_field is None:
        incident_angle_field = get_field_name('incident_angle')
    if sigma_0_field is None:
        sigma_0_field = get_field_name('sigma_0')
    if effective_area_field is None:
        effective_area_field = get_field_name('effective_area')
    if rcs_clutter_field is None:
        rcs_clutter_field = get_field_name('rcs_clutter')
    if dBm_clutter_field is None:
        dBm_clutter_field = get_field_name('dBm_clutter')
    if dBZ_clutter_field is None:
        dBZ_clutter_field = get_field_name('dBZ_clutter')
    if visibility_polar_field is None:
        visibility_polar_field = get_field_name('visibility_polar')

    elevations   = radar.fixed_angle['data']
    azimuths_pol = radar.get_azimuth(0)
    range_pol    = radar.range['data']

    # Define aeqd projection for radar local Cartesian coords
    pargs = pyproj.Proj(proj="aeqd", lat_0 = radar.latitude['data'][0],
                 lon_0 = radar.longitude['data'][0], datum = "WGS84",
                 units="m")

    # Define coordinate transform: (local radar Cart coords) -> (DEM coords)
    transformer = pyproj.Transformer.from_proj(pargs, dem_grid.projection)

    # Get local radar coordinates at elevaiton = 0
    xr, yr, _ = antenna_vectors_to_cartesian(radar.range['data'],
                          radar.get_azimuth(0), 0, ke = ke)

    # Project them in DEM proj
    xr_proj, yr_proj = transformer.transform(xr,yr)
    
    # Get local radar coordinates at elevaiton = 0
    xr, yr, _ = antenna_vectors_to_cartesian(radar.range['data'],
                          radar.get_azimuth(0), 0, ke = ke)
    
    # Project them in DEM proj
    xr_proj, yr_proj = transformer.transform(xr,yr)
    rad_x = xr_proj[0,0] # radar x coord in DEM proj
    rad_y = yr_proj[0,0] # radar y coord in DEM proj
    
    # Clip DEM outside radar domain
    if clip:
        dem_grid = gf.clip_grid(dem_grid, xr_proj, yr_proj)

    res_dem = dem_grid.metadata['resolution']
    xmin_dem = np.min(dem_grid.x['data'])
    ymin_dem = np.min(dem_grid.y['data'])

    # Processing starts here...
    ###########################################################################
    # 1) Compute range map
    logging.info('1) computing radar range map...')
    X_dem, Y_dem = np.meshgrid(dem_grid.x['data'], dem_grid.y['data'])
    range_map = np.sqrt((X_dem - rad_x) ** 2 + (Y_dem - rad_y) ** 2)

    # 2) Compute azimuth map
    logging.info('2) computing radar azimuth map...')

    az_map = (np.arctan2((X_dem - rad_x), (Y_dem - rad_y)) + 2*np.pi) % (2*np.pi)
    az_map *= 180 / np.pi

    # 3) Compute bent DEM map
    logging.info('3) computing bent DEM...')
    # Index first level
    dem = np.ma.filled(dem_grid.fields['terrain_altitude']['data'], np.nan)[0]
    _, _, zgrid = antenna_to_cartesian(range_map / 1000., az_map, 0, ke = ke)
    bent_map = dem - (zgrid + radar.altitude['data'])

    # 4) Compute slope and aspect
    logging.info('4) computing DEM slope and aspect...')
    gx = sobel(dem, axis = 1) / (8 * res_dem) # gradient w-e direction
    gy = sobel(dem, axis = 0) /  (8 * res_dem) # gradient s-n direction
    slope_map = np.arctan(np.sqrt(gy**2 + gx**2)) * 180 / np.pi
    aspect_map = (np.arctan2(gy, -gx) + np.pi) * 180 / np.pi

    # 5) Compute theta (elevation) angle at topography
    logging.info('5) computing radar elevation angle at topography...')
    elev_map = (np.arctan2(bent_map, range_map) * 180 / np.pi)

    # 6) COmpute visibility map and minimum visible elevation angle map
    logging.info('6) computing radar visibility map...')
    visib_map, minviselev_map = gf.visibility(az_map, range_map, elev_map,
                res_dem, xmin_dem, ymin_dem, rad_x, rad_y, dr, daz, verbose)

    # 7) Compute min visible altitude
    logging.info('7) computing min visible altitude...')
    R = 6371.0 * 1000.0 * ke     # effective radius of earth in meters.
    minvisalt_map = ((range_map ** 2 + R ** 2 + 2.0 * range_map * R *
           np.sin((minviselev_map + radar_specs['beamwidth'] / 2.) *
                  np.pi / 180.)) ** 0.5 - R)

    # 8) Compute effective area
    logging.info('8) computing effective area...')
    effarea_map = res_dem ** 2 / np.cos(slope_map * np.pi / 180.0)

    # 9) Compute local incidence angle
    logging.info('9) computing local incidence angle...')
    slope  = slope_map* np.pi / 180.0
    aspect = aspect_map * np.pi / 180.0
    zenith = (90. - elev_map * np.pi / 180.0 )
    az     = az_map* np.pi / 180.0

    incang_map = np.arccos(-( np.sin(slope) * np.sin(zenith) *
          (np.sin(aspect) * np.sin(az) + np.cos(aspect) * np.cos(az)) +
                  np.cos(slope) * np.cos(zenith))) * 180 / np.pi

    # 10) Compute sigma 0
    logging.info('10) computing sigma0...')
    sigma0_map = gf.sigma0(incang_map, radar_specs['frequency'],
                           sigma0_method)

    # Processing for every elevation angle starts here...
    ###########################################################################
    # 11) Compute rcs
    strelevs = ','.join([str(e) for e in elevations])
    logging.info('13) computing polar RCS at elevations {:s}...'
                     .format(strelevs))

    rcs_pol =  gf.rcs(az_map, range_map, elev_map, effarea_map, sigma0_map,
                         visib_map, range_pol, azimuths_pol, elevations,
                         res_dem, xmin_dem, ymin_dem, rad_x, rad_y,
                         radar_specs['beamwidth'], radar_specs['tau'],
                         raster_oversampling, verbose)
    rcs_pol = np.ma.array(rcs_pol, mask = np.isnan(rcs_pol),
                         fill_value = fill_value)

    # 12) Clutter power map in dBm
    logging.info('12) computing clutter power in dBm...')
    range_pol_e = np.tile(range_pol,(rcs_pol.shape[0], 1))
    range_log = 10* np.log10(range_pol_e)
    sigma_map = 10 * np.log10(rcs_pol)

    lambd = 3. / (radar_specs['frequency'] * 10.)
    pconst = (10 * np.log10(radar_specs['power']) + 2 * radar_specs['gain'] +
              20 * np.log10(lambd) -  radar_specs['loss']
              - 30 * np.log10(4 * np.pi))

    clutter_dBm_pol = (pconst - 4 * range_log -
                   2 * atm_att * range_pol_e / 1000. + sigma_map)

    # 13) Clutter reflectivity map in dBZ
    logging.info('13) computing clutter reflectivity in dBZ...')
    lambd = 3. / (radar_specs['frequency'] * 10.)
    dbzconst = (10 * np.log10(16 * np.log(2)) + 40 * np.log10(lambd) -
                10 * np.log10(radar_specs['tau'] * 3e8) -
                20 * np.log10(radar_specs['beamwidth'] * np.pi / 180.) -
                60 * np.log10(np.pi) - 20 * np.log10(mosotti_kw))

    convert_dbzm_to_dbz = 180. # 10*log10(1 m^6 / 1 mm^6) = 180
    clutter_dBZ_pol = (sigma_map - 2 * range_log + dbzconst +
                       convert_dbzm_to_dbz)

    # 14) Visibility map by angle
    logging.info('13) computing polar visibility at elevations {:s}...'.
                     format(strelevs))

    vispol =  gf.visibility_angle(minviselev_map, az_map, range_map,
                                  range_pol, azimuths_pol, elevations,
                                  res_dem, xmin_dem, ymin_dem, rad_x, rad_y,
                                  radar_specs['beamwidth'], radar_specs['tau'],
                                  az_conv, raster_oversampling, verbose)
    vispol = np.ma.array(vispol, mask = np.isnan(vispol),
                         fill_value = fill_value)

    logging.info('All done, creating outputs...')

    # Note that the [None,:,:]  indexing is for compatibility with
    # pyart grid.add_field() function
    bent_terrain_altitude_dic = get_metadata(bent_terrain_altitude_field)
    bent_terrain_altitude_dic['data'] = bent_map[None,:,:]

    elevation_dic = get_metadata(elevation_angle_field)
    elevation_dic['data'] = elev_map[None,:,:]

    terrain_slope_dic = get_metadata(terrain_slope_field)
    terrain_slope_dic['data'] = slope_map[None,:,:]

    terrain_aspect_dic = get_metadata(terrain_aspect_field)
    terrain_aspect_dic['data'] = aspect_map[None,:,:]

    visibility_dic = get_metadata(visibility_field)
    visibility_dic['data'] = visib_map[None,:,:]

    min_vis_elevation_dic = get_metadata(min_vis_elevation_field)
    min_vis_elevation_dic['data'] = minviselev_map[None,:,:]

    min_vis_altitude_dic = get_metadata(min_vis_altitude_field)
    min_vis_altitude_dic['data'] = minvisalt_map[None,:,:]

    incident_angle_dic = get_metadata(incident_angle_field)
    incident_angle_dic['data'] = incang_map[None,:,:]

    effective_area_dic = get_metadata(effective_area_field)
    effective_area_dic['data'] = effarea_map[None,:,:]

    sigma_0_dic = get_metadata(sigma_0_field)
    sigma_0_dic['data'] = sigma0_map[None,:,:]

    rcs_clutter_dic = get_metadata(rcs_clutter_field)
    rcs_clutter_dic['data'] = rcs_pol

    dBm_clutter_dic = get_metadata(dBm_clutter_field)
    dBm_clutter_dic['data'] = clutter_dBm_pol

    dBZ_clutter_dic = get_metadata(dBZ_clutter_field)
    dBZ_clutter_dic['data'] = clutter_dBZ_pol

    visibility_polar_dic = get_metadata(visibility_polar_field)
    visibility_polar_dic['data'] = vispol

    if not return_pyart_objects:
        return (bent_terrain_altitude_dic, terrain_slope_dic,
                terrain_aspect_dic, elevation_dic, min_vis_elevation_dic,
                min_vis_altitude_dic, visibility_dic, incident_angle_dic,
                effective_area_dic, sigma_0_dic, rcs_clutter_dic,
                dBm_clutter_dic, dBZ_clutter_dic, visibility_polar_dic)

    logging.info('Adding Cartesian output fields to input Grid (DEM) object...')
    dem_grid.add_field(bent_terrain_altitude_field,
                       bent_terrain_altitude_dic)
    dem_grid.add_field(elevation_angle_field,
                       elevation_dic)
    dem_grid.add_field(terrain_slope_field,
                       terrain_slope_dic)
    dem_grid.add_field(terrain_aspect_field,
                       terrain_aspect_dic)
    dem_grid.add_field(visibility_field,
                       visibility_dic)
    dem_grid.add_field(min_vis_elevation_field,
                       min_vis_elevation_dic)
    dem_grid.add_field(min_vis_altitude_field,
                       min_vis_altitude_dic)
    dem_grid.add_field(incident_angle_field,
                       incident_angle_dic)
    dem_grid.add_field(effective_area_field,
                       effective_area_dic)
    dem_grid.add_field(sigma_0_field,
                       sigma_0_dic)

    logging.info('Creating Radar (DEM) object...')
    new_radar = deepcopy(radar)
    new_radar.fields = dict()

    radar.add_field(rcs_clutter_field,
                       rcs_clutter_dic)
    radar.add_field(dBm_clutter_field,
                       dBm_clutter_dic)
    radar.add_field(dBZ_clutter_field,
                       dBZ_clutter_dic)
    radar.add_field(visibility_polar_field,
                       visibility_polar_dic)

    return dem_grid, radar




# THIS FILE IS GENERATED FROM PYART SETUP.PY
short_version = '0.6.5'
version = '0.6.5'
full_version = '0.6.5.dev+72fd6cf'
git_revision = '72fd6cf806a6abdd5c17bfcf54a7740a55eb0a4c'
release = False

if not release:
    version = full_version



import numpy as np


def _steiner_conv_strat(refl, x, y, dx, dy, intense=42, peak_relation=0,
                        area_relation=1, bkg_rad=11000, use_intense=True):
    """
    We perform the Steiner et al. (1995) algorithm for echo classification
    using only the reflectivity field in order to classify each grid point
    as either convective, stratiform or undefined. Grid points are classified
    as follows,

    0 = Undefined
    1 = Stratiform
    2 = Convective
    """
    def convective_radius(ze_bkg, area_relation):
        """
        Given a mean background reflectivity value, we determine via a step
        function what the corresponding convective radius would be

        Higher background reflectivitives are expected to have larger convective
        influence on surrounding areas, so a larger convective radius would be
        prescribed
        """
        if (area_relation == 0):
            if (ze_bkg < 30):
                conv_rad = 1000.
            elif (ze_bkg >= 30) & (ze_bkg < 35.):
                conv_rad = 2000.
            elif (ze_bkg >= 35.) & (ze_bkg < 40.):
                conv_rad = 3000.
            elif (ze_bkg >= 40.) & (ze_bkg < 45.):
                conv_rad = 4000.
            else:
                conv_rad = 5000.

        if (area_relation == 1):
            if (ze_bkg < 25):
                conv_rad = 1000.
            elif (ze_bkg >= 25) & (ze_bkg < 30.):
                conv_rad = 2000.
            elif (ze_bkg >= 30.) & (ze_bkg < 35.):
                conv_rad = 3000.
            elif (ze_bkg >= 35.) & (ze_bkg < 40.):
                conv_rad = 4000.
            else:
                conv_rad = 5000.

        if (area_relation == 2):
            if (ze_bkg < 20):
                conv_rad = 1000.
            elif (ze_bkg >= 20) & (ze_bkg < 25.):
                conv_rad = 2000.
            elif (ze_bkg >= 25.) & (ze_bkg < 30.):
                conv_rad = 3000.
            elif (ze_bkg >= 30.) & (ze_bkg < 35.):
                conv_rad = 4000.
            else:
                conv_rad = 5000.

        if (area_relation == 3):
            if (ze_bkg < 40):
                conv_rad = 0.
            elif (ze_bkg >= 40) and (ze_bkg < 45.):
                conv_rad = 1000.
            elif (ze_bkg >= 45.) and (ze_bkg < 50.):
                conv_rad = 2000.
            elif (ze_bkg >= 50.) and (ze_bkg < 55.):
                conv_rad = 6000.
            else:
                conv_rad = 8000.

        return conv_rad

    def peakedness(ze_bkg, peak_relation):
        """
        Given a background reflectivity value, we determine what the necessary
        peakedness (or difference) has to be between a grid point's reflectivity
        and the background reflectivity in order for that grid point to be labeled
        convective
        """
        if (peak_relation == 0):
            if (ze_bkg < 0.):
                peak = 10.
            elif (ze_bkg >= 0.) and (ze_bkg < 42.43):
                peak = 10. - ze_bkg ** 2 / 180.
            else:
                peak = 0.

        elif (peak_relation == 1):
            if (ze_bkg < 0.):
                peak = 14.
            elif (ze_bkg >= 0.) and (ze_bkg < 42.43):
                peak = 14. - ze_bkg ** 2 / 180.
            else:
                peak = 4.

        return peak

    sclass = np.zeros(refl.shape, dtype=int)
    ny, nx = refl.shape

    for i in range(0, nx):
        # Get stencil of x grid points within the background radius
        imin = np.max(np.array([1, (i - bkg_rad / dx)], dtype=int))
        imax = np.min(np.array([nx, (i + bkg_rad / dx)], dtype=int))

        for j in range(0, ny):
            # First make sure that the current grid point has not already been classified.
            # This can happen when grid points within the convective radius of a previous
            # grid point have also been classified
            if ~np.isnan(refl[j, i]) & (sclass[j, i] == 0):
                # Get stencil of y grid points within the background radius
                jmin = np.max(np.array([1, (j - bkg_rad / dy)], dtype=int))
                jmax = np.min(np.array([ny, (j + bkg_rad / dy)], dtype=int))

                n = 0
                sum_ze = 0

                # Calculate the mean background reflectivity for the current grid point,
                # which will be used to determine the convective radius and the required peakedness

                for l in range(imin, imax):
                    for m in range(jmin, jmax):
                        rad = np.sqrt((x[l] - x[i]) ** 2 + (y[m] - y[j]) ** 2)

                        # The mean background reflectivity will first be computed in
                        # linear units, i.e. mm^6/m^3, then converted to decibel units.
                        if rad <= bkg_rad:
                            n += 1
                            sum_ze += 10. ** (refl[m, l] / 10.)

                ze_bkg = 10.0 * np.log10(sum_ze / n)

                # Now get the corresponding convective radius knowing the mean background reflectivity
                conv_rad = convective_radius(ze_bkg, area_relation)

                # Now we want to investigate the points surrounding the current
                # grid point that are within the convective radius, and whether
                # they too are convective, stratiform or undefined

                # Get stencil of x and y grid points within the convective radius
                lmin = np.max(np.array([1, int(i - conv_rad / dx)], dtype=int))
                lmax = np.min(np.array([nx, int(i + conv_rad / dx)], dtype=int))
                mmin = np.max(np.array([1, int(j - conv_rad / dy)], dtype=int))
                mmax = np.min(np.array([ny, int(j + conv_rad / dy)], dtype=int))

                if use_intense and (refl[j, i] >= intense):
                    sclass[j, i] = 2

                    for l in range(lmin, lmax):
                        for m in range(mmin, mmax):
                            if not np.isnan(refl[j, i]):
                                rad = np.sqrt((x[l] - x[i]) ** 2 + (y[m] - y[j]) ** 2)

                                if rad <= conv_rad:
                                    sclass[m, l] = 2

                else:
                    peak = peakedness(ze_bkg, peak_relation)

                    if refl[j, i] - ze_bkg >= peak:
                        sclass[j, i] = 2

                        for l in range(imin, imax):
                            for m in range(jmin, jmax):
                                if not np.isnan(refl[j, i]):
                                    rad = np.sqrt((x[l] - x[i]) ** 2 + (y[m] - y[j]) ** 2)

                                    if rad <= conv_rad:
                                        sclass[m, l] = 2

                    else:
                        # If by now the current grid point has not been classified as convective
                        # by either the intensity criteria or the peakedness criteria, then it must be stratiform
                        sclass[j, i] = 1

    return sclass


def steiner_class_buff(ze, x, y, z, dx, dy, bkg_rad,
                       work_level, intense, peak_relation,
                       area_relation, use_intense,):

    zslice = np.argmin(np.abs(z - work_level))
    refl = ze[zslice, :, :]

    area_rel = {"small": 0, "medium": 1, "large": 2, "sgp": 3}
    peak_rel = {"default": 0, "sgp": 1}

    sclass = _steiner_conv_strat(refl, x, y, dx, dy, intense=intense,
                                 peak_relation=peak_rel[peak_relation],
                                 area_relation=area_rel[area_relation],
                                 bkg_rad=11000, use_intense=True)

    return sclass



import pyart
import matplotlib.pyplot as plt
import numpy as np
from pyrad.io import read_data_mxpol

rhi = read_data_mxpol.pyrad_MXPOL(pyart.testing.RHI_ML_FILE,
                                  field_names = ['Zh','Rhohv'])

# Remove 

refl_field = 'reflectivity'
rhohv_field = 'uncorrected_cross_correlation_ratio'

ml = pyart.retrieve.detect_ml(rhi, refl_field = refl_field,
                       rhohv_field = rhohv_field, 
                       interp_holes = False, 
                       max_length_holes = 250)

plt.figure()
plt.plot(ml[0]['ml_cart']['x'],ml[0]['ml_cart']['bottom_ml'])
plt.plot(ml[0]['ml_cart']['x'],ml[0]['ml_cart']['top_ml'])
plt.xlabel('Distance from radar [m]')
plt.ylabel('Height above radar [m]')
plt.legend(['ML bottom','ML top'])
plt.show()
